# GromacsWrapper: gw-partial_tempering.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from gromacs.scaling import partial_tempering
import argparse

description = """
Modify gromacs processed topology (processed.top, generated by `grompp -pp` command)
for running with solute tempering replica exchange (REST2). 
"""


def parse_args():
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--scale_protein",
        type=float,
        help="scale protein interactions by this scaling factor (0-1)",
    )
    parser.add_argument(
        "--scale_lipids",
        type=float,
        help="scale lipid interactions by this scaling factor (0-1)",
    )
    parser.add_argument(
        "--scale_protein_lipids",
        type=float,
        help="scale lipid-lipid interactions by this scaling factor (0-1)",
    )
    parser.add_argument("input", help="input topology (processed.top)")
    parser.add_argument("output", help="output topology")
    parser.add_argument(
        "--banned_lines",
        default="",
        help="line numbers of dihedrals/impropers that one wishes to exclude from scaling",
    )
    return parser.parse_args()


args = parse_args()
partial_tempering(args)



# GromacsWrapper: test_example.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import pytest

import gromacs


common_tool_names = ["pdb2gmx", "grompp", "editconf", "mdrun"]
aliased_tool_names = list(gromacs.tools.NAMES5TO4.values())

changed_tools = {
    "do_dssp": "dssp",  # GMX 2022.x --> 2023.x
}


@pytest.fixture(scope="function", params=set(common_tool_names + aliased_tool_names))
def gromacs_tool(request):
    # for testing, work around changed tools that cannot be aliased
    # via NAMES5TO4 because the call signature has changed: replace
    # old toolname with the functionally equivalent new name
    try:
        tool = getattr(gromacs, request.param)
    except AttributeError:
        tool = getattr(gromacs, changed_tools[request.param])
    return tool


def test_tools_help(gromacs_tool):
    rc, out, err = gromacs_tool(h=True, stdout=False, stderr=False)
    assert rc == 0, "Gromacs command {0} failed".format(gromacs_tool.command_name)
    assert out + err, "Gromacs command {0} produced no output for -h".format(
        gromacs_tool.command_name
    )


def test_failure_raises():
    # unknown option
    with pytest.raises(gromacs.GromacsError):
        gromacs.grompp(y=True)


def test_failure_warns():
    # unknown option
    grompp_warn = gromacs.tools.Grompp(failure="warn")
    with pytest.warns(gromacs.GromacsFailureWarning):
        grompp_warn(y=True)


def test_failure_ignore():
    # unknown option
    grompp_ignore = gromacs.tools.Grompp(failure=None)
    try:
        grompp_ignore(y=True)
    except Exception as err:
        raise AssertionError("Should have ignored exception {}".format(err))


class TestRelease(object):
    # add tested releases here
    major_releases = ("4", "5", "2016", "2018", "2019", "2020", "2021", "2022", "2023")

    def test_release(self):
        assert gromacs.release().startswith(self.major_releases)

    def test_release_startswith(self):
        assert gromacs.release.startswith(self.major_releases)

    def test_str(self):
        assert str(gromacs.release()) == gromacs.release()



# GromacsWrapper: test_charmm22st.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import pytest

from .top import TopologyTest
from ...datafiles import datafile


class TestCharmm22st(TopologyTest):
    processed = datafile("fileformats/top/charmm22st/processed.top")
    conf = datafile("fileformats/top/charmm22st/conf.gro")
    molecules = ["SOL", "Protein", "Ion", "Cal", "Ces", "CL", "K", "NA", "ZN"]



# GromacsWrapper: test_example.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

import gromacs
import pytest
from numpy.testing import assert_equal

from ..datafiles import datafile


@pytest.fixture(params=["original", "nofilename", "written"])
def SIMPLE_NDX(request, tmpdir):
    ndx = gromacs.fileformats.ndx.NDX(datafile("simple.ndx"))

    if request.param == "written":
        out = str(tmpdir.join("out.ndx"))

        ndx.write(out)
        ndx = gromacs.fileformats.ndx.NDX(out)
    elif request.param == "nofilename":
        ndx = gromacs.fileformats.ndx.NDX()
        ndx.read(datafile("simple.ndx"))

    return ndx


def test_read(SIMPLE_NDX):
    ndx = SIMPLE_NDX

    assert_equal(ndx["Oxygen"], [1, 4, 7])
    assert_equal(ndx["Hydrogen"], [2, 3, 5, 6, 8, 9])


def test_get(SIMPLE_NDX):
    assert_equal(SIMPLE_NDX.get("Oxygen"), [1, 4, 7])


def test_set(SIMPLE_NDX):
    SIMPLE_NDX["Nitrogen"] = [10, 11, 12]

    assert_equal(SIMPLE_NDX["Nitrogen"], [10, 11, 12])


def test_size(SIMPLE_NDX):
    assert len(SIMPLE_NDX) == 2


def test_sizes(SIMPLE_NDX):
    assert SIMPLE_NDX.sizes == {"Oxygen": 3, "Hydrogen": 6}


def test_groups(SIMPLE_NDX):
    assert list(SIMPLE_NDX.groups) == ["Oxygen", "Hydrogen"]



# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.
""":mod:`gromacs.tools` -- Gromacs commands classes
================================================

The underlying idea of GromacsWrapper is to automatically generate Python
classes that run the actual GROMACS tools. These classes are written in such a
way that they can be called as if they were functions with parameters
resembling the GROMACS tools commandline arguments. This a two-step process
when :mod:`gromacs` is imported: First a tool class is generated for each
GROMACS tool (the :ref:`command-list`). Then an *instance* of each of these
classes is instantiated in placed into the top level :mod:`gromacs`
module. Thus, :class:`gromacs.grompp` is an instance of
:class:`gromacs.tools.Grompp`. Users typically only access the tool *instances*
at the top level of the module.

.. note::

   Because the tool instances are autogenerated on the fly and depend on the
   installed GROMACS release, there is no autogenerated documentation available
   for them. Use in Python ``help(gromacs.grompp)`` to get help.

The following documentation for the :mod:`gromacs.tools` module describes in
more detail how GROMACS tools are generated and managed and is primarily of
interest to developers.


Gromacs tool instantiation
--------------------------

A Gromacs command class produces an instance of a Gromacs tool command
(:class:`gromacs.core.GromacsCommand`), any argument or keyword argument
supplied will be used as default values for when the command is run.

Classes have the same name of the corresponding Gromacs tool with the first
letter capitalized and dot and dashes replaced by underscores to make it a
valid python identifier.

The list of tools to be loaded is configured with the ``tools`` and ``groups``
options of the ``~/.gromacswrapper.cfg`` file. Guesses are made if these
options are not provided; see :mod:`gromacs.config` for details.

In the following example we create two instances of the
:class:`gromacs.tools.Trjconv` command (which runs the Gromacs ``trjconv``
command)::

  from gromacs.tools import Trjconv

  trjconv = tools.Trjconv()
  trjconv_compact = tools.Trjconv(ur='compact', center=True, boxcenter='tric', pbc='mol',
                                  input=('protein','system'))

The first one, ``trjconv``, behaves as the standard commandline tool but the
second one, ``trjconv_compact``, will by default create a compact
representation of the input data by taking into account the shape of the unit
cell. Of course, the same effect can be obtained by providing the corresponding
arguments to ``trjconv`` but by naming the more specific command differently
one can easily build up a library of small tools that will solve a specific,
repeatedly encountered problem reliably. This is particularly helpful when doing
interactive work.


Aliased commands
----------------

GromacsWrapper has been around since ancient GROMACS 4.5.x and throughout the
history it has provided a way to run different versions of GROMACS commands
with the same Python script in a backwards compatible manner by *aliasing*
equivalent GROMACS tools to the same GromacsWrapper tool (using
:data:`NAMES5TO4`). Modern GROMACS (since 5.x and throughout 2016-2023) tools
are aliased to their Gromacs 4 tool names for backwards compatibility.

For example, in "classic" GROMACS we had the :command:`g_sas` command that
became :command:`gmx sasa` since GROMACS 5.x. In GromacsWrapper you can access
the same tool as :class:`gromacs.g_sas` or :class:`gromacs.sasa`.

.. warning::

   You should check that the different GROMACS versions of tools give
   equivalent answers for your problem. The aliasing just makes it easy for you
   to call the tool in the same manner. You are still responsible for
   validating your own results.

Sometimes GROMACS changes commands in an way that is fundamentally incompatible
and in this way there is not much that GromacsWrapper can do. The best you can
probably do in your own scripts is to use :class:`gromacs.release` (see docs
for :class:`gromacs.tools.Release`) to check in your own script which release
of GROMACS is running and then call different tools depending on what you
find. For example, GROMACS 2023 replaced :command:`gmx do_dssp` with
:command:`gmx dssp` which is not directly argument-compatible so GromacsWrapper
does *not* alias it. Therefore, scripts relying on :class:`gromacs.do_dssp`
will break unless you account for it explicitly with code similar to ::

  release_year = int(gromacs.release()[:4])
  if release_year >= 2023:
      gromacs.dssp(s=TPR, f=XTC, sel="Protein",
                   o="dssp.dat", pbc=True, hmode="dssp")
  else:
      gromacs.do_dssp(s=TPR, f=XTC, input=["Protein"]
                      ssdump="ssdump.dat", o="ss.xpm", sc="scount.xvg")

.. note::

   It is recommended to keep a single version of all tools for a project and
   record the version in the methods section of a publication.



Multi index
-----------

It is possible to extend the tool commands and patch in additional
functionality. For example, the :class:`GromacsCommandMultiIndex` class makes a
command accept multiple index files and concatenates them on the fly; the
behaviour mimics Gromacs' "multi-file" input that has not yet been enabled for
all tools.

.. autoclass:: GromacsCommandMultiIndex
.. autofunction:: merge_ndx


.. _virtual-gromacs-commands:

Virtual Gromacs commands
------------------------

The following "commands" do not exist as tools in the Gromacs package
but are added here because they are useful.

.. autoclass:: Release


Helpers
-------

These helper functions are necessary for collecting and setting up the
Gromacs tools. They are mostly of interest to developers.

.. autodata:: NAMES5TO4
.. autodata:: V4TOOLS
.. autofunction:: tool_factory
.. autofunction:: load_v4_tools
.. autofunction:: load_v5_tools
.. autofunction:: find_executables
.. autofunction:: make_valid_identifier
.. autoexception:: GromacsToolLoadingError

Gromacs tools
-------------

Each command class in the :ref:`command-list` below is used to create
a command instance in the top level :mod:`gromacs` module *if the
Gromacs tools can be found in the file system* (see
:mod:`gromacs.config`). For example, the class :class:`Grompp` is used
to create the command :class:`gromacs.grompp`.


Registry
~~~~~~~~

The :ref:`command-list` below reflects the Gromacs commands that were
available when the documentation was built and can vary from
installation to installation. All currently available Gromacs commands
are listed in the dictionary :data:`gromacs.tools.registry`, in
particular, ``gromacs.tools.registry.keys()`` lists the names.

.. data:: registry
   :annotation: = {'Grompp': <gromacs.tools.Grompp', 'Mdrun': <gromacs.tools.Mdrun', ...}

   :class:`dict` that contains all currently available Gromacs
   commands as well as the :ref:`virtual-gromacs-commands`. The
   :data:`registry` is generated when the :mod:`gromacs.tools` package
   is imported for the first time.


.. _command-list:

Command list
~~~~~~~~~~~~

The list below reflects the Gromacs commands that were available when
the documentation was built and can vary from installation to
installation. All currently available Gromacs commands are listed in
the dictionary :data:`gromacs.tools.registry`, which can be processed
at run time.

.. The following is autogenerated with sphinx.

"""
from __future__ import absolute_import

import six

import os.path
import tempfile
import subprocess
import atexit
import logging
import re

from . import config
from .core import GromacsCommand

logger = logging.getLogger("gromacs.tools")

#: List of tools coming with standard Gromacs 4.x.
V4TOOLS = (
    "g_cluster",
    "g_dyndom",
    "g_mdmat",
    "g_principal",
    "g_select",
    "g_wham",
    "mdrun",
    "do_dssp",
    "g_clustsize",
    "g_enemat",
    "g_membed",
    "g_protonate",
    "g_sgangle",
    "g_wheel",
    "mdrun_d",
    "editconf",
    "g_confrms",
    "g_energy",
    "g_mindist",
    "g_rama",
    "g_sham",
    "g_x2top",
    "mk_angndx",
    "eneconv",
    "g_covar",
    "g_filter",
    "g_morph",
    "g_rdf",
    "g_sigeps",
    "genbox",
    "pdb2gmx",
    "g_anadock",
    "g_current",
    "g_gyrate",
    "g_msd",
    "g_sorient",
    "genconf",
    "g_anaeig",
    "g_density",
    "g_h2order",
    "g_nmeig",
    "g_rms",
    "g_spatial",
    "genion",
    "tpbconv",
    "g_analyze",
    "g_densmap",
    "g_hbond",
    "g_nmens",
    "g_rmsdist",
    "g_spol",
    "genrestr",
    "trjcat",
    "g_angle",
    "g_dielectric",
    "g_helix",
    "g_nmtraj",
    "g_rmsf",
    "g_tcaf",
    "gmxcheck",
    "trjconv",
    "g_bar",
    "g_dih",
    "g_helixorient",
    "g_order",
    "g_rotacf",
    "g_traj",
    "gmxdump",
    "trjorder",
    "g_bond",
    "g_dipoles",
    "g_kinetics",
    "g_pme_error",
    "g_rotmat",
    "g_tune_pme",
    "grompp",
    "g_bundle",
    "g_disre",
    "g_lie",
    "g_polystat",
    "g_saltbr",
    "g_vanhove",
    "make_edi",
    "xpm2ps",
    "g_chi",
    "g_dist",
    "g_luck",
    "g_potential",
    "g_sas",
    "g_velacc",
    "make_ndx",
)


#: dict of names in Gromacs 5 that correspond to an equivalent tool in
#: in Gromacs 4. The names are literal Gromacs names.
NAMES5TO4 = {
    # same name in both versions
    "grompp": "grompp",
    "eneconv": "eneconv",
    "editconf": "editconf",
    "pdb2gmx": "pdb2gmx",
    "trjcat": "trjcat",
    "trjconv": "trjconv",
    "trjorder": "trjorder",
    "xpm2ps": "xpm2ps",
    "mdrun": "mdrun",
    "make_ndx": "make_ndx",
    "make_edi": "make_edi",
    "genrestr": "genrestr",
    "genion": "genion",
    "genconf": "genconf",
    "do_dssp": "do_dssp",  # removed in GMX 2023, replaced with native `gmx dssp`
    # changed names
    "convert-tpr": "tpbconv",
    "dump": "gmxdump",
    "check": "gmxcheck",
    "solvate": "genbox",
    "distance": "g_dist",
    "sasa": "g_sas",
    "gangle": "g_sgangle",
}


class GromacsToolLoadingError(Exception):
    """Raised when no Gromacs tool could be found."""


class GromacsCommandMultiIndex(GromacsCommand):
    """Command class that accept multiple index files.

    It works combining multiple index files into a single temporary one so
    that tools that do not (yet) support multi index files as input can be
    used as if they did.

    It creates a new file only if multiple index files are supplied.
    """

    def __init__(self, **kwargs):
        kwargs = self._fake_multi_ndx(**kwargs)
        super(GromacsCommandMultiIndex, self).__init__(**kwargs)

    def run(self, *args, **kwargs):
        kwargs = self._fake_multi_ndx(**kwargs)
        return super(GromacsCommandMultiIndex, self).run(*args, **kwargs)

    def _fake_multi_ndx(self, **kwargs):
        ndx = kwargs.get("n")
        if (
            not (ndx is None or isinstance(ndx, six.string_types))
            and len(ndx) > 1
            and "s" in kwargs
        ):
            ndx.append(kwargs.get("s"))
            kwargs["n"] = merge_ndx(*ndx)
        return kwargs


def tool_factory(clsname, name, driver, base=GromacsCommand):
    """Factory for GromacsCommand derived types."""
    clsdict = {
        "command_name": name,
        "driver": driver,
        "__doc__": property(base._get_gmx_docs),
    }
    return type(clsname, (base,), clsdict)


def make_valid_identifier(name):
    """Turns tool names into valid identifiers.

    :param name: tool name
    :return: valid identifier
    """
    return name.replace("-", "_").capitalize()


def find_executables(path):
    """Find executables in a path.

    Searches executables in a directory excluding some know commands
    unusable with GromacsWrapper.

    :param path: dirname to search for
    :return: list of executables
    """
    execs = []
    for exe in os.listdir(path):
        fullexe = os.path.join(path, exe)
        if (
            os.access(fullexe, os.X_OK)
            and not os.path.isdir(fullexe)
            and exe
            not in [
                "GMXRC",
                "GMXRC.bash",
                "GMXRC.csh",
                "GMXRC.zsh",
                "demux.pl",
                "xplor2gmx.pl",
            ]
        ):
            execs.append(exe)
    return execs


def load_v5_tools():
    """Load Gromacs 2023/.../2016/5.x tools automatically using some heuristic.

    Tries to load tools (1) using the driver from configured groups (2) and
    falls back to automatic detection from ``GMXBIN`` (3) then to rough guesses.

    In all cases the command ``gmx help`` is run to get all tools available.

    :return: dict mapping tool names to GromacsCommand classes
    """
    logger.debug("Loading 2023/.../2016/5.x tools...")

    drivers = config.get_tool_names()

    if len(drivers) == 0 and "GMXBIN" in os.environ:
        drivers = find_executables(os.environ["GMXBIN"])

    if len(drivers) == 0 or len(drivers) > 4:
        drivers = ["gmx", "gmx_d", "gmx_mpi", "gmx_mpi_d"]

    append = config.cfg.getboolean("Gromacs", "append_suffix", fallback=True)

    tools = {}
    for driver in drivers:
        suffix = driver.partition("_")[2]
        try:
            out = subprocess.check_output([driver, "-quiet", "help", "commands"])
            for line in out.splitlines()[5:-1]:
                line = str(
                    line.decode("ascii")
                )  # Python 3: byte string -> str, Python 2: normal string

                if len(line) > 4:
                    if (line[4] != " ") and (" " in line[4:]):
                        name = line[4 : line.index(" ", 4)]
                        fancy = make_valid_identifier(name)
                        if suffix and append:
                            fancy = "{0!s}_{1!s}".format(fancy, suffix)
                        tools[fancy] = tool_factory(fancy, name, driver)
        except (subprocess.CalledProcessError, OSError):
            pass

    if not tools:
        errmsg = "Failed to load 2023/.../2016/5.x tools (tried drivers: {})".format(
            drivers
        )
        logger.debug(errmsg)
        raise GromacsToolLoadingError(errmsg)
    logger.debug("Loaded {0} v5 tools successfully!".format(len(tools)))
    return tools


def load_v4_tools():
    """Load Gromacs 4.x tools automatically using some heuristic.

    Tries to load tools (1) in configured tool groups (2) and fails back  to
    automatic detection from ``GMXBIN`` (3) then to a prefilled list.

    Also load any extra tool configured in ``~/.gromacswrapper.cfg``

    :return: dict mapping tool names to GromacsCommand classes
    """
    logger.debug("Loading v4 tools...")

    names = config.get_tool_names()

    if len(names) == 0 and "GMXBIN" in os.environ:
        names = find_executables(os.environ["GMXBIN"])

    if len(names) == 0 or len(names) > len(V4TOOLS) * 4:
        names = list(V4TOOLS)

    names.extend(config.get_extra_tool_names())

    tools = {}
    for name in names:
        fancy = make_valid_identifier(name)
        tools[fancy] = tool_factory(fancy, name, None)

    if not tools:
        errmsg = "Failed to load v4 tools"
        logger.debug(errmsg)
        raise GromacsToolLoadingError(errmsg)
    logger.debug("Loaded {0} v4 tools successfully!".format(len(tools)))
    return tools


def merge_ndx(*args):
    """Takes one or more index files and optionally one structure file and
    returns a path for a new merged index file.

    :param args: index files and zero or one structure file
    :return: path for the new merged index file
    """
    ndxs = []
    struct = None
    for fname in args:
        if fname.endswith(".ndx"):
            ndxs.append(fname)
        else:
            if struct is not None:
                raise ValueError("only one structure file supported")
            struct = fname

    fd, multi_ndx = tempfile.mkstemp(suffix=".ndx", prefix="multi_")
    os.close(fd)
    atexit.register(os.unlink, multi_ndx)

    if struct:
        make_ndx = registry["Make_ndx"](f=struct, n=ndxs, o=multi_ndx)
    else:
        make_ndx = registry["Make_ndx"](n=ndxs, o=multi_ndx)

    _, _, _ = make_ndx(input=["q"], stdout=False, stderr=False)
    return multi_ndx


# Load tools
if config.MAJOR_RELEASE in (
    "5",
    "2016",
    "2018",
    "2019",
    "2020",
    "2021",
    "2022",
    "2023",
):
    logger.debug(
        "Trying to load configured Gromacs major release {0}".format(
            config.MAJOR_RELEASE
        )
    )
    registry = load_v5_tools()
elif config.MAJOR_RELEASE == "4":
    logger.debug(
        "Trying to load configured Gromacs major release {0}".format(
            config.MAJOR_RELEASE
        )
    )
    registry = load_v4_tools()
else:
    logger.debug("No major release configured: trying 2023/.../2016/5.x -> 4.x")
    try:
        registry = load_v5_tools()
    except GromacsToolLoadingError:
        try:
            registry = load_v4_tools()
        except GromacsToolLoadingError:
            errmsg = "Autoloading was unable to load any Gromacs tool"
            logger.critical(errmsg)
            raise GromacsToolLoadingError(errmsg)

# Aliases command names to run unmodified GromacsWrapper scripts on a machine
# with only 5.x
# update with temporary directory
for fancy, cmd in list(registry.items()):
    for c5, c4 in six.iteritems(NAMES5TO4):
        # have to check each one, since it's possible there are suffixes
        # like for double precision; cmd.command_name is Gromacs name
        # (e.g. 'convert-tpr') so we need to be careful in the processing below.
        name = cmd.command_name
        if name.startswith(c5):
            if c4 == c5:
                break
            else:
                # maintain suffix (note: need to split with fancy because Gromacs
                # names (c5) may contain '-' etc)
                name = c4 + fancy.split(make_valid_identifier(c5))[1]
                registry[make_valid_identifier(name)] = registry[fancy]
                break
    else:
        # the common case of just adding the 'g_'
        registry["G_{0!s}".format(fancy.lower())] = registry[fancy]


# Patching up commands that may be useful to accept multiple index files
for name4, name5 in [("G_mindist", "Mindist"), ("G_dist", "Distance")]:
    if name4 in registry:
        cmd = registry[name4]
        registry[name4] = tool_factory(
            name4, cmd.command_name, cmd.driver, GromacsCommandMultiIndex
        )
        if name5 in registry:
            registry[name5] = registry[name4]


# create a release "virtual command" (issue #161)
class Release(object):
    """Release string of the currently loaded Gromacs version.

    :return:   str,
        Release string such as "2018.2" or "4.6.5" or ``None``
        if Gromacs can not be found.

    .. Note::

       The release string is obtained from the output of ``gmx grompp
       -version``, specifically the line starting with ``Gromacs
       version:``. If this changes then this function will break.

    .. rubric:: Example

    This command allows user code to work around known issues with
    old/new versions of Gromacs::

      if gromacs.release.startswith("4"):
        # do something for classic Gromacs
      else:
        # do it the modern way

    (Note that calling ``gromacs.release()`` will simply return the
    release string, which is equivalent to
    ``str(gromacs.release)``. For convenience, the :meth:`startswith`
    method exists, which directly works with the release string.)


    .. versionadded:: 0.8.0

    """

    gromacs_version = re.compile(
        "^G[rR][oO][mM][aA][cC][sS] version:" "\s*(VERSION)?\s*(?P<version>.+)$"
    )

    def __init__(self):
        self.release = None
        try:
            grompp = registry["Grompp"]()
            rc, out, err = grompp(version=True, stdout=False, stderr=False)
            output_lines = out.splitlines() + err.splitlines()
        except KeyError:
            output_lines = []

        for line in output_lines:
            line = line.strip()
            m = self.gromacs_version.match(line)
            if m:
                self.release = m.group("version")
                break

    def __call__(self):
        if self.release is None:
            logger.warning("gromacs.release(): cannot determine Gromacs release")

        return self.release

    def startswith(self, *args, **kwargs):
        """gromacs.release.startswith(prefix[, start[, end]]) -> bool

        Return ``True`` if the Gromacs release string starts with the
        specified `prefix`, ``False`` otherwise.  With optional
        `start`, test the string beginning at that position.  With
        optional `end`, stop comparing the release string at that
        position.  `prefix` can also be a tuple of strings to try, for
        example ::

          gromacs.release.startswith(('2016', '2018', '2019'))

        """
        return self.release.startswith(*args, **kwargs)

    def __str__(self):
        return self.release or "unknown"


# Append class doc for each command
for name in six.iterkeys(registry):
    __doc__ += ".. class:: {0!s}\n    :noindex:\n".format(name)

registry["Release"] = Release

# Finally add command classes to module's scope
globals().update(registry)
__all__ = ["GromacsCommandMultiIndex", "merge_ndx"]
__all__.extend(list(registry.keys()))



# GromacsWrapper: test_example.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

import gromacs
import pytest
from numpy.testing import assert_equal

from ..datafiles import datafile


@pytest.fixture(
    params=["original", "written", "no_autoconvert"],
)
def CUSTOM_EM_MDP(request, tmpdir):
    autoconvert = not request.param == "no_autoconvert"
    mdp = gromacs.fileformats.mdp.MDP(
        datafile("custom_em.mdp"), autoconvert=autoconvert
    )
    if request.param == "written":
        # to check that written mdp has same data
        out = str(tmpdir.join("out.mdp"))
        mdp.write(out)
        mdp = gromacs.fileformats.mdp.MDP(out)
    return mdp, autoconvert


class TestMDP(object):
    def test_values(self, CUSTOM_EM_MDP):
        mdp, autoconvert = CUSTOM_EM_MDP

        if autoconvert:

            def conv(val):
                return val

        else:

            def conv(val):
                return str(val)

        if autoconvert:
            assert_equal(mdp["include"], ["-I.", "-I..", "-I../top"])
            assert_equal(mdp["define"], "-DPOSRES")
        else:
            assert mdp["include"] == "-I. -I.. -I../top"
            assert mdp["define"] == "-DPOSRES"
        assert mdp["integrator"] == "cg"
        assert mdp["emtol"] == conv(500)
        assert mdp["emstep"] == conv(0.01)
        assert mdp["nsteps"] == conv(1000)
        assert mdp["nstcgsteep"] == conv(100)
        assert mdp["constraints"] == "none"
        assert mdp["nstcomm"] == conv(1)
        assert mdp["cutoff-scheme"] == "Verlet"
        assert mdp["vdwtype"] == "cutoff"
        assert mdp["coulombtype"] == "PME"
        assert mdp["ns_type"] == "grid"
        assert mdp["rlist"] == conv(1.0)
        assert mdp["rcoulomb"] == conv(1.0)
        assert mdp["rvdw"] == conv(1.0)
        assert mdp["rvdw-switch"] == conv(0.8)
        assert mdp["Tcoupl"] == "no"
        assert mdp["Pcoupl"] == "no"
        assert mdp["gen_vel"] == "no"
        assert mdp["nstxout"] == conv(0)

    def test_comments(self, CUSTOM_EM_MDP):
        mdp, _ = CUSTOM_EM_MDP

        assert mdp["C0001"] == "custom EM"
        assert sum(1 for k in mdp if k.startswith("C0")) == 1

    def test_blank_lines(self, CUSTOM_EM_MDP):
        mdp, _ = CUSTOM_EM_MDP
        assert sum(1 for k in mdp if k.startswith("B0")) == 6

    def test_no_filename(self, tmpdir):
        mdp = gromacs.fileformats.mdp.MDP()

        mdp["define"] = ["-DPOSRES", "-DTHIS"]
        mdp["vdwtype"] = "cutoff"
        mdp["nsteps"] = 1234

        out = str(tmpdir.join("out.mdp"))
        mdp.write(out)

        back = gromacs.fileformats.mdp.MDP(out)

        assert_equal(back["define"], ["-DPOSRES", "-DTHIS"])
        assert back["vdwtype"] == "cutoff"
        assert back["nsteps"] == 1234


@pytest.fixture
def NONSENSE_MDP(tmpdir):
    outfile = str(tmpdir.join("nonsense.mdp"))

    with open(datafile("custom_em.mdp"), "r") as infile:
        data = infile.read()

    data += "errors: plenty\n"

    with open(outfile, "w") as out:
        out.write(data)

    return outfile


def test_bad_mdp(NONSENSE_MDP):
    with pytest.raises(
        gromacs.ParseError, match="unknown line in mdp file, 'errors: plenty'"
    ):
        gromacs.fileformats.mdp.MDP(NONSENSE_MDP)



# GromacsWrapper: test_example.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import contextlib
import os
import errno
import sys

if sys.version_info[0] < 3:
    from ConfigParser import NoOptionError, NoSectionError
else:
    from configparser import NoOptionError, NoSectionError

import pytest

import gromacs.config
import gromacs.utilities


@pytest.fixture
def GMXRC():
    # Try using GMXRC in config file:
    GMXRC = gromacs.config.cfg.get("Gromacs", "gmxrc")
    if GMXRC:
        return GMXRC
    # get GMXRC from installed Gromacs conda package
    for gmxexe in ("gmx", "gmx_d", "gmx_mpi", "gmx_mpi_d", "grompp", "mdrun"):
        path = gromacs.utilities.which(gmxexe)
        if path is not None:
            break
    else:
        raise RuntimeError("Cannot find Gromacs installation")

    bindir = os.path.dirname(path)
    GMXRC = os.path.join(bindir, "GMXRC")
    if not os.path.exists(GMXRC):
        raise IOError(errno.ENOENT, "Could not find Gromacs setup file", GMXRC)
    return GMXRC


@contextlib.contextmanager
def temp_environ():
    _environ = os.environ.copy()
    try:
        yield os.environ
    finally:
        os.environ.clear()
        os.environ.update(_environ)


def test_set_gmxrc_environment(GMXRC):
    # not threadsafe: function modifies the global process environment

    gmx_envvars = (
        "GMXBIN",
        "GMXLDLIB",
        "GMXMAN",
        "GMXDATA",
        "GMXPREFIX",
        "GROMACS_DIR",
    )

    with temp_environ() as environ:
        # clean environment so that we can detect changes
        for envvar in gmx_envvars:
            try:
                del environ[envvar]
            except KeyError:
                pass

        before = environ.copy()
        gromacs.config.set_gmxrc_environment(GMXRC)
        newvars = set(environ) - set(before)
        for envvar in gmx_envvars:
            assert envvar in newvars, "GMX environment variable was not added correctly"


def test_check_setup():
    rc = gromacs.config.check_setup()
    assert rc in (True, False)


def test_get_configuration():
    cfg = gromacs.config.get_configuration()
    # could test more variables
    assert cfg.getpath("DEFAULT", "configdir")
    assert isinstance(cfg.getboolean("Gromacs", "append_suffix"), bool)


def test_modified_config(modified_config):
    tools, append_suffix, Path = modified_config
    if tools != "":
        assert Path("~/gmx_mpi").expanduser().exists()
    assert gromacs.config.cfg.get("Gromacs", "tools") == tools
    assert gromacs.config.cfg.get("Gromacs", "append_suffix") == append_suffix


def test_get_boolean():
    # The code for getboolean was custom implemented based on the Python 3.7
    # ConfigParser.getboolean code.
    # These tests should be unnecessary for the Python 3 version of the code.
    cfg = gromacs.config.cfg
    assert isinstance(cfg.getboolean("Gromacs", "append_suffix"), bool)
    assert isinstance(cfg.getboolean("Gromacs", "append_suffix", fallback=True), bool)
    with pytest.raises(ValueError):
        cfg.getboolean("DEFAULT", "configdir")
    with pytest.raises(ValueError):
        cfg.getboolean("DEFAULT", "configdir", fallback=True)
    with pytest.raises(NoOptionError):
        cfg.getboolean("Gromacs", "bool")
    with pytest.raises(NoSectionError):
        cfg.getboolean("Not a section", "bool")
    cfg.set("Gromacs", "bool", "")
    cfg.remove_option("Gromacs", "bool")
    assert cfg.getboolean("Gromacs", "bool", fallback=True) is True
    with pytest.raises(NoOptionError):
        cfg.getboolean("Gromacs", "bool")



import os.path
import pytest

import gromacs.qsub


def test_queuing_systems(known=("Sun Gridengine", "PBS", "LoadLeveler", "Slurm")):
    assert len(gromacs.qsub.queuing_systems) == len(known)
    for qs in gromacs.qsub.queuing_systems:
        assert qs.name in known
        assert qs.flag("-x").endswith("-x")
        assert qs.isMine("script.sge") in (True, False)
        directories = ["run1", "run2", "run3"]
        try:
            assert qs.array(directories)
        except NotImplementedError:
            pass


@pytest.mark.parametrize(
    "scriptfile,name",
    [
        ("foo.sge", "Sun Gridengine"),
        ("foo.pbs", "PBS"),
        ("foo.ll", "LoadLeveler"),
        ("foo.slu", "Slurm"),
    ],
)
def test_detect_queuing_system(scriptfile, name):
    qs = gromacs.qsub.detect_queuing_system(scriptfile)
    assert qs.name == name


def test_generate_submit_scripts(tmpdir):
    templates = ["local.sh", "darwin.sh"]
    with tmpdir.as_cwd():
        names = gromacs.qsub.generate_submit_scripts(templates)
        for name in names:
            assert os.path.exists(name)
    basenames = [os.path.basename(name) for name in names]
    assert basenames == templates

    # todo: check utf-8...



import pytest
import logging

import gromacs
import gromacs.log

import logging

# This is almost certainly not thread/parallel safe.


@pytest.fixture
def logger_with_file(tmp_path):
    logfile = tmp_path / "gmx.log"
    logger = gromacs.log.create("GMX", logfile=str(logfile))
    return logfile, logger


def test_create(logger_with_file):
    logfile, logger = logger_with_file
    logger.info("Jabberwock")
    logger.debug("Cheshire Cat")
    txt = logfile.read_text()
    assert "Jabberwock" in txt
    assert "Cheshire Cat" in txt


def test_clear_handlers(logger_with_file):
    logfile, logger = logger_with_file
    gromacs.log.clear_handlers(logger)
    logger.warning("Dodo")
    txt = logfile.read_text()
    assert "Dodo" not in txt


def test_NullHandler():
    h = gromacs.log.NullHandler()
    logger = logging.getLogger("GMX")
    logger.addHandler(h)
    logger.warning("screaming in silence")
    assert True  # not sure what to test here


@pytest.fixture
def gromacs_logger(tmp_path):
    logfile = tmp_path / "gromacs.log"
    gromacs.start_logging(logfile=str(logfile))
    logger = logging.getLogger("gromacs")
    logger.info("Running a test for logging")
    gromacs.stop_logging()
    return logfile


def _assert_msg_in_log(logfile, msg):
    output = logfile.read_text()
    assert msg in output


def test_start_logger(gromacs_logger):
    _assert_msg_in_log(
        gromacs_logger, "GromacsWrapper {} STARTED".format(gromacs.__version__)
    )


def test_using_logger(gromacs_logger):
    _assert_msg_in_log(gromacs_logger, "Running a test for logging")


def test_stop_logger(gromacs_logger):
    _assert_msg_in_log(
        gromacs_logger, "GromacsWrapper {} STOPPED".format(gromacs.__version__)
    )



from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib

import pytest
from numpy.testing import assert_almost_equal, assert_equal

import gromacs.fileformats.xvg
from gromacs.formats import XVG


@pytest.fixture(scope="module")
def data():
    data = np.random.normal(loc=2.1, scale=0.5, size=(6, 1000))
    data[0] = 0.1 * np.arange(data.shape[1])
    return data


@pytest.fixture(scope="module")
def correldata(omega=2 * np.pi / 5):
    t = np.linspace(-100, 100, 10000)
    Y1 = 1 * np.sin(omega * t) - 4
    Y2 = -0.5 * np.sin(0.3 * omega * t)
    DY = np.random.normal(scale=2, size=(2, len(t)))
    return np.vstack([t, Y1 + DY[0], Y2 + DY[1]])


class TestXVG_array(object):
    @pytest.fixture
    def xvg(self, data):
        return XVG(array=data.copy(), names="t,a,b,c,d,e")

    def test_names(self, xvg):
        assert_equal(xvg.names, ["t", "a", "b", "c", "d", "e"])

    def test_array(self, xvg, data):
        assert_equal(xvg.array.shape, data.shape)
        assert_almost_equal(xvg.array, data)

    @pytest.mark.parametrize("name", ("mean", "max", "min", "std"))
    def test_props(self, xvg, data, name):
        assert_almost_equal(getattr(xvg, name), getattr(data[1:], name)(axis=1))

    def test_write_read(self, xvg, tmpdir):
        fname = "random.xvg"
        with tmpdir.as_cwd():
            xvg.write(fname)
            newxvg = XVG(filename=fname)
        assert_almost_equal(newxvg.array, xvg.array)
        ## will fail: column names are not written
        # assert_equal(newxvg.names, xvg.names)

    def test_correl(self, correldata):
        xvg = XVG(array=correldata, names="t,y1,y2")
        # FIXME
        # force=True : TypeError: tcorrel() got an unexpected keyword argument 'force'
        xvg.set_correlparameters(nstep=None, ncorrel=25000)

        sigma = xvg.error
        tc = xvg.tc
        assert_equal(sigma.shape, (2,))
        assert_equal(tc.shape, (2,))

    @pytest.mark.parametrize(
        "method",
        ("mean", "circmean", "min", "max", "rms", "percentile", "smooth", "error"),
    )
    def test_decimate(self, correldata, method, maxpoints=100):
        xvg = XVG(array=correldata)
        data = xvg.array
        reduced = xvg.decimate(method, data, maxpoints=maxpoints)
        assert_equal(reduced.shape, (len(data), maxpoints))

    def test_plot(self, xvg):
        ax = xvg.plot()
        assert isinstance(ax, matplotlib.axes.Axes)

    def test_errorplot(self, xvg):
        ax = xvg.errorbar(maxpoints=1000)
        assert isinstance(ax, matplotlib.axes.Axes)

    def test_plot_coarsend(self, xvg):
        ax = xvg.plot_coarsened(maxpoints=100)
        assert isinstance(ax, matplotlib.axes.Axes)


def test_break_array():
    angles = np.pi * np.array([-1.9, -1, -1, -0.5, 0, 0.9, 1.5, 2, -2, -1.4])
    expected = np.pi * np.array([-1.9, -1, -1, -0.5, 0, 0.9, 1.5, 2, np.NAN, -2, -1.4])
    other = np.ones_like(angles)
    ma, mother = gromacs.fileformats.xvg.break_array(
        angles, threshold=np.pi, other=other
    )
    assert isinstance(ma, np.ma.core.MaskedArray)
    assert isinstance(mother, np.ma.core.MaskedArray)
    assert_almost_equal(ma, expected)
    assert len(ma) == len(mother)



from __future__ import print_function

from gromacs.fileformats import TOP
import numpy as np
import math
import copy, argparse


def scale_angles(mol, angles):
    new_angles = {}
    for dh in mol.angles:
        atypes = (
            dh.atom1.get_atomtype(),
            dh.atom2.get_atomtype(),
            dh.atom3.get_atomtype(),
        )
        atypes = [a.replace("_", "").replace("=", "") for a in atypes]
        for iswitch in range(16):
            if iswitch % 2 == 0:
                a1 = atypes[0]
                a2 = atypes[1]
                a3 = atypes[2]
            else:
                a1 = atypes[2]
                a2 = atypes[1]
                a3 = atypes[0]

            if (iswitch // 2) % 2 == 1:
                a1 = "X"
            if (iswitch // 4) % 2 == 1:
                a2 = "X"
            if (iswitch // 8) % 2 == 1:
                a3 = "X"
            key = "{0}-{1}-{2}-{3}".format(a1, a2, a3, dh.gromacs["func"])
            if key in angles:
                for i, at in enumerate(angles[key]):
                    # new_angles.append(at)
                    new_angles[key] = at
                break
    return new_angles.values()


def scale_dihedrals(mol, dihedrals):
    new_dihedrals = {}
    for dh in mol.dihedrals:
        atypes = (
            dh.atom1.get_atomtype(),
            dh.atom2.get_atomtype(),
            dh.atom3.get_atomtype(),
            dh.atom4.get_atomtype(),
        )
        atypes = [a.replace("_", "").replace("=", "") for a in atypes]
        for iswitch in range(32):
            if iswitch % 2 == 0:
                a1 = atypes[0]
                a2 = atypes[1]
                a3 = atypes[2]
                a4 = atypes[3]
            else:
                a1 = atypes[3]
                a2 = atypes[2]
                a3 = atypes[1]
                a4 = atypes[0]

            if (iswitch // 2) % 2 == 1:
                a1 = "X"
            if (iswitch // 4) % 2 == 1:
                a2 = "X"
            if (iswitch // 8) % 2 == 1:
                a3 = "X"
            if (iswitch // 16) % 2 == 1:
                a4 = "X"
            key = "{0}-{1}-{2}-{3}-{4}".format(a1, a2, a3, a4, dh.gromacs["func"])
            if key in dihedrals:
                for i, dt in enumerate(dihedrals[key]):
                    # new_dihedrals.append(dt)
                    new_dihedrals[key] = dt
                break

    print(new_dihedrals)
    return new_dihedrals.values()


def scale_impropers(mol, impropers):
    new_impropers = {}
    for im in mol.impropers:
        atypes = (
            im.atom1.get_atomtype(),
            im.atom2.get_atomtype(),
            im.atom3.get_atomtype(),
            im.atom4.get_atomtype(),
        )
        atypes = [a.replace("_", "").replace("=", "") for a in atypes]
        for iswitch in range(32):
            if iswitch % 2 == 0:
                a1 = atypes[0]
                a2 = atypes[1]
                a3 = atypes[2]
                a4 = atypes[3]
            else:
                a1 = atypes[3]
                a2 = atypes[2]
                a3 = atypes[1]
                a4 = atypes[0]
            if (iswitch / 2) % 2 == 1:
                a1 = "X"
            if (iswitch / 4) % 2 == 1:
                a2 = "X"
            if (iswitch / 8) % 2 == 1:
                a3 = "X"
            if (iswitch / 16) % 2 == 1:
                a4 = "X"
            key = "{0}-{1}-{2}-{3}-{4}".format(a1, a2, a3, a4, im.gromacs["func"])
            if key in impropers:
                for i, imt in enumerate(impropers[key]):
                    new_impropers[key] = imt
                break
    print(new_impropers)
    return new_impropers.values()


parser = argparse.ArgumentParser()
parser.add_argument("input")
parser.add_argument("output")
args = parser.parse_args()

top = TOP(args.input)
molname = top.molecules[0].name
mol = top.dict_molname_mol[molname]

#
# ATOMTYPES
#
atomtypes = {a.atomtype for a in mol.atoms}
top.atomtypes = [at for at in top.atomtypes if at.atype in atomtypes]

#
# BONDTYPES
#
bondtypes = {tuple(sorted((b.atom1.atomtype, b.atom2.atomtype))) for b in mol.bonds}
bondtypes_dictionary = {
    tuple(sorted((bt.atype1, bt.atype2))): bt for bt in top.bondtypes
}
top.bondtypes = [bondtypes_dictionary[bt] for bt in bondtypes]

#
# Build bond dictionary
#
angletypes = {}
for at in top.angletypes:
    name = "{0}-{1}-{2}-{3}".format(at.atype1, at.atype2, at.atype3, at.gromacs["func"])
    if not name in angletypes:
        angletypes[name] = []
    angletypes[name].append(at)

#
# Build dihedral dictionary
#
dihedraltypes = {}
for dt in top.dihedraltypes:
    name = "{0}-{1}-{2}-{3}-{4}".format(
        dt.atype1, dt.atype2, dt.atype3, dt.atype4, dt.gromacs["func"]
    )
    if not name in dihedraltypes:
        dihedraltypes[name] = []
    dihedraltypes[name].append(dt)
print("Build dihedraltypes dictionary with {0} entries".format(len(dihedraltypes)))

#
# Build improper dictionary
#
impropertypes = {}
for it in top.impropertypes:
    name = "{0}-{1}-{2}-{3}-{4}".format(
        it.atype1, it.atype2, it.atype3, it.atype4, it.gromacs["func"]
    )
    if not name in impropertypes:
        impropertypes[name] = []
    impropertypes[name].append(it)
print("Build impropertypes dictionary with {0} entries".format(len(impropertypes)))

top.angletypes = scale_angles(mol, angletypes)
top.dihedraltypes = scale_dihedrals(mol, dihedraltypes)
top.impropertypes = scale_impropers(mol, impropertypes)

top.nonbond_params = []
top.cmaptypes = []
atomtypes = {at.atype for at in top.atomtypes}

pairtypes = [
    pt for pt in top.pairtypes if (pt.atype1 in atomtypes) and (pt.atype2 in atomtypes)
]
top.pairtypes = pairtypes

# Remove non-default moleculetypes
for k in top.dict_molname_mol.keys():
    if k in [molname]:
        continue
    del top.dict_molname_mol[k]

top.write(args.output)



# GromacsWrapper: test_example.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import os.path
import re

import pytest
from numpy.testing import assert_almost_equal

import gromacs
from gromacs import cbook
import gromacs.setup

from .datafiles import datafile


@pytest.fixture
def simulation(tmpdir):
    pdb = datafile("1ake_A.pdb")
    with tmpdir.mkdir("top").as_cwd():
        f = gromacs.setup.topology(struct=pdb, ff="oplsaa", water="tip4p")
        yield f


@pytest.mark.xfail(
    gromacs.release.startswith("2020.6"),
    reason="pdb2gmx 2020.6 fails to build the TIP4P waters",
)
def test_grompp_qtot(tmpdir, simulation):
    with tmpdir.mkdir("qtot").as_cwd():
        with open("none.mdp", "w") as mdp:
            mdp.write("; empty mdp file\nrcoulomb = 1\nrvdw = 1\nrlist = 1\n")
        qtot = cbook.grompp_qtot(
            f="none.mdp",
            c=simulation["struct"],
            p=simulation["top"],
            stdout=False,
            maxwarn=10,
        )
    assert_almost_equal(
        qtot,
        -4,
        decimal=5,
        err_msg="grompp_qtot() failed to compute total charge correctly",
    )


@pytest.mark.xfail(
    gromacs.release.startswith("2020.6"),
    reason="pdb2gmx 2020.6 fails to build the TIP4P waters",
)
def test_portable_topology(tmpdir, simulation):
    with tmpdir.mkdir("processed").as_cwd():
        pptopol = cbook.create_portable_topology(
            simulation["top"], simulation["struct"]
        )

    # correct filename
    assert os.path.split(pptopol)[-1].startswith("pp_")

    lines = open(pptopol).readlines()
    subsections_only = [line.strip() for line in lines if line.startswith("[")]

    assert re.match(r";\s+File 'system.top' was generated", lines[1])
    assert re.match(r";\s+This is a standalone topology file", lines[6])
    for subsection in (
        "[ defaults ]",
        "[ atomtypes ]",
        "[ bondtypes ]",
        "[ constrainttypes ]",
        "[ angletypes ]",
        "[ dihedraltypes ]",
        "[ moleculetype ]",
        "[ atoms ]",
        "[ bonds ]",
        "[ pairs ]",
        "[ angles ]",
        "[ dihedrals ]",
        "[ system ]",
        "[ molecules ]",
    ):
        assert subsection in subsections_only



# GromacsWrapper: test_amber03star.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import pytest

import gromacs

from .top import TopologyTest
from ...datafiles import datafile


class TestAmber03star(TopologyTest):
    processed = datafile("fileformats/top/amber03star/processed.top")
    conf = datafile("fileformats/top/amber03star/conf.gro")
    molecules = [
        "Protein",
        "SOL",
        "IB+",
        "CA",
        "CL",
        "NA",
        "MG",
        "K",
        "RB",
        "CS",
        "LI",
        "ZN",
    ]

    @pytest.mark.xfail(
        gromacs.release().startswith(("2022", "2023")),
        reason="issue #236 https://github.com/Becksteinlab/GromacsWrapper/issues/236",
    )
    def test_mdrun(self, tmpdir, low_performance):
        super(TestAmber03star, self).test_mdrun(tmpdir, low_performance)



# GromacsWrapper: top.py
# Copyright (c) 2012 Oliver Beckstein <orbeckst@gmail.com>
# Copyright (c) 2010 Tsjerk Wassenaar <tsjerkw@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

"""
Gromacs TOP file format
=======================

Classes
-------
.. autoclass:: TOP
   :members:

.. autoclass:: SystemToGroTop
   :members:


History
-------

Sources adapted from code by Reza Salari https://github.com/resal81/PyTopol


Example: Read a processed.top file and scale charges
----------------------------------------------------

Run ``grompp -pp`` to produce a processed.top from conf.gro, grompp.mdp and topol.top files::

  $ grompp -pp

This file now containts all the force-field information::

  from gromacs.fileformats import TOP
  top = TOP("processed.top")

Scale the LJ epsilon by an arbitrary number, here 0.9 ::

  scaling = 0.9
  for at in top.atomtypes:
    at.gromacs['param']['lje'] *= scaling

Write out the scaled down topology::

  top.write("output.top")

.. Note::

   You can use this to prepare a series of top files for Hamiltonian Replica
   Exchange (HREX) simulations. See ``scripts/gw-partial_tempering.py`` for an example.

"""
from __future__ import absolute_import

import textwrap
import logging
from collections import OrderedDict as odict

from . import blocks


class TOP(blocks.System):
    """Class to make a TOP object from a GROMACS processed.top file

    The force-field and molecules data is exposed as python object.

    .. Note::

       Only processed.top files generated by GROMACS 'grompp -pp'
       are supported - the usual topol.top files are not supported (yet!)

    """

    default_extension = "top"
    logger = logging.getLogger("gromacs.fileformats.TOP")

    def __init__(self, fname):
        """Initialize the TOP structure.

        :Arguments:
          *fname*
              name of the processed.top file
        """
        super(TOP, self).__init__()

        self.fname = fname

        self.defaults = {
            "nbfunc": None,
            "comb-rule": None,
            "gen-pairs": None,
            "fudgeLJ": None,
            "fudgeQQ": None,
        }

        self.dict_molname_mol = odict()  # contains molname:mol
        self.found_sections = []
        self.forcefield = "gromacs"

        self.molecules = []
        self._parse(fname)
        self.molecules = tuple(self.molecules)

    def write(self, filename):
        """Write the TOP object to a file"""
        SystemToGroTop(self, filename)

    def __repr__(self):
        """Represent the TOP object as a string"""
        moltypenames = list(self.dict_molname_mol.keys())
        moltypenames.sort()

        data = []
        data.append("\n")

        main_items = set(
            ["atomtypes", "pairtypes", "bondtypes", "angletypes", "dihedraltypes"]
        )
        other_items = [
            "{0:s} ({1:d})".format(m, len(self.information[m]))
            for m in list(self.information.keys())
            if m not in main_items
        ]
        other_items = " ".join(other_items)
        nattype = len(self.atomtypes)
        nprtype = len(self.pairtypes)
        nbndtype = len(self.bondtypes)
        nangtype = len(self.angletypes)
        ndihtype = len(self.dihedraltypes)
        nimptype = len(self.impropertypes)
        data.append(
            "{0:>20s}  {1:>7s} {2:>7s} {3:>7s} {4:>7s} {5:>7s} {6:>7s}".format(
                "Param types:", "atom", "pair", "bond", "ang", "dih", "imp"
            )
        )
        msg = "{0:20s}  {1:7d} {2:7d} {3:7d} {4:7d} {5:7d} {6:7d}    {7:s}".format(
            "", nattype, nprtype, nbndtype, nangtype, ndihtype, nimptype, other_items
        )
        data.append("=" * 69)
        data.append(msg)
        data.append("\n")

        main_items = set(["atoms", "pairs", "bonds", "angles", "dihedrals"])
        data.append(
            "{0:>20s}  {1:>7s} {2:>7s} {3:>7s} {4:>7s} {5:>7s} {6:>7s}".format(
                "Params:", "atom", "pair", "bond", "ang", "dih", "imp"
            )
        )
        data.append("=" * 69)
        for mname in moltypenames:
            mol = self.dict_molname_mol[mname]
            other_items = [
                "{0:s} ({1:d})".format(m, len(mol.information[m]))
                for m in list(mol.information.keys())
                if m not in main_items
            ]
            other_items = " ".join(other_items)

            natoms = len(mol.atoms)
            npairs = len(mol.pairs)
            nbonds = len(mol.bonds)
            nangles = len(mol.angles)
            ndih = len(mol.dihedrals)
            nimp = len(mol.impropers)
            msg = "{0:20s}  {1:7d} {2:7d} {3:7d} {4:7d} {5:7d} {6:7d}    {7:s}".format(
                mol.name, natoms, npairs, nbonds, nangles, ndih, nimp, other_items
            )
            data.append(msg)

        return "\n".join(data)

    def _parse(self, fname):
        """Parse a processed.top GROMACS topology file

        The function reads in the file line-by-line, and it's a bunch of 'elif' statements,
        writing parameter/atom line to current section/molecule.

        ParamTypes are added to self.xyztypes (AtomType goes to self.atomtypes).
        Params are added to current molecule (Atom goes to mol.atoms.append(atom))

        MoleculeTypes and Molecules are odd, and are added to
            * MoleculeType to :attr:`self.dict_molname_mol[mol.name] = mol`
            * Molecule to :attr:`self.molecules.append(self.dict_molname_mol[mname])`

        :obj:`curr_sec` variable stores to current section being read-in
        :obj:`mol` variable stores the current molecule being read-in
        :obj:`cmap_lines` are a little odd, since CMAP parameters are stored on multiple lines

        :Arguments:
          *fname*
              name of the processed.top file

        :Returns: None
        """

        def _find_section(line):
            return line.strip("[").strip("]").strip()

        def _add_info(sys_or_mol, section, container):
            # like (mol, 'atomtypes', mol.atomtypes)
            if sys_or_mol.information.get(section, False) is False:
                sys_or_mol.information[section] = container

        mol = None  # to hold the current mol
        curr_sec = None
        cmap_lines = []

        with open(fname) as f:
            for i_line, line in enumerate(f):
                # trimming
                if ";" in line:
                    line = line[0 : line.index(";")]
                line = line.strip()

                if line == "":
                    continue

                if line[0] == "*":
                    continue

                # the topology must be stand-alone (i.e. no includes)
                if line.startswith("#include"):
                    msg = 'The topology file has "#include" statements.'
                    msg += " You must provide a processed topology file that grompp creates."
                    raise ValueError(msg)

                # find sections
                if line[0] == "[":
                    curr_sec = _find_section(line)
                    self.found_sections.append(curr_sec)
                    continue

                fields = line.split()

                if curr_sec == "defaults":
                    """
                    # ; nbfunc        comb-rule       gen-pairs       fudgeLJ fudgeQQ
                    #1               2               yes             0.5     0.8333
                    """
                    assert len(fields) in [2, 5]
                    self.defaults["nbfunc"] = int(fields[0])
                    self.defaults["comb-rule"] = int(fields[1])
                    if len(fields) == 5:
                        self.defaults["gen-pairs"] = fields[2]
                        self.defaults["fudgeLJ"] = float(fields[3])
                        self.defaults["fudgeQQ"] = float(fields[4])

                elif curr_sec == "atomtypes":
                    """
                    # ;name               at.num    mass         charge    ptype  sigma   epsilon
                    # ;name   bond_type   at.num    mass         charge    ptype  sigma   epsilon
                    # ;name                         mass         charge    ptype  c6      c12

                    """
                    if len(fields) not in (6, 7, 8):
                        self.logger.warning(
                            "skipping atomtype line with neither 7 or 8 fields: \n {0:s}".format(
                                line
                            )
                        )
                        continue

                    # shift = 0 if len(fields) == 7 else 1
                    shift = len(fields) - 7
                    at = blocks.AtomType("gromacs")
                    at.atype = fields[0]
                    if shift == 1:
                        at.bond_type = fields[1]

                    at.mass = float(fields[2 + shift])
                    at.charge = float(fields[3 + shift])

                    particletype = fields[4 + shift]
                    assert particletype in ("A", "S", "V", "D")
                    if particletype not in ("A",):
                        self.logger.warning(
                            'warning: non-atom particletype: "{0:s}"'.format(line)
                        )

                    sig = float(fields[5 + shift])
                    eps = float(fields[6 + shift])

                    at.gromacs = {
                        "param": {"lje": eps, "ljl": sig, "lje14": None, "ljl14": None}
                    }

                    self.atomtypes.append(at)

                    _add_info(self, curr_sec, self.atomtypes)

                # extend system.molecules
                elif curr_sec == "moleculetype":
                    assert len(fields) == 2

                    mol = blocks.Molecule()

                    mol.name = fields[0]
                    mol.exclusion_numb = int(fields[1])

                    self.dict_molname_mol[mol.name] = mol

                elif curr_sec == "atoms":
                    """
                    #id    at_type     res_nr  residu_name at_name  cg_nr  charge   mass  typeB    chargeB      massB
                    # 1       OC          1       OH          O1       1      -1.32

                    OR

                    [ atoms ]
                    ; id   at type  res nr  residu name at name     cg nr   charge
                    1       OT      1       SOL              OW             1       -0.834

                    """

                    aserial = int(fields[0])
                    atype = fields[1]
                    resnumb = int(fields[2])
                    resname = fields[3]
                    aname = fields[4]
                    cgnr = int(fields[5])
                    charge = float(fields[6])
                    rest = fields[7:]

                    atom = blocks.Atom()
                    atom.name = aname
                    atom.atomtype = atype
                    atom.number = aserial
                    atom.resname = resname
                    atom.resnumb = resnumb
                    atom.charge = charge

                    if rest:
                        mass = float(rest[0])
                        atom.mass = mass

                    mol.atoms.append(atom)

                    _add_info(mol, curr_sec, mol.atoms)

                elif curr_sec in ("pairtypes", "pairs", "pairs_nb"):
                    """
                    section     #at     fu      #param
                    ---------------------------------
                    pairs       2       1       V,W
                    pairs       2       2       fudgeQQ, qi, qj, V, W
                    pairs_nb    2       1       qi, qj, V, W

                    """

                    ai, aj = fields[:2]
                    fu = int(fields[2])
                    assert fu in (1, 2)

                    pair = blocks.InteractionType("gromacs")
                    if fu == 1:
                        if curr_sec == "pairtypes":
                            pair.atype1 = ai
                            pair.atype2 = aj
                            v, w = list(map(float, fields[3:5]))
                            pair.gromacs = {
                                "param": {
                                    "lje": None,
                                    "ljl": None,
                                    "lje14": w,
                                    "ljl14": v,
                                },
                                "func": fu,
                            }

                            self.pairtypes.append(pair)
                            _add_info(self, curr_sec, self.pairtypes)

                        elif curr_sec == "pairs":
                            ai, aj = list(map(int, [ai, aj]))
                            pair.atom1 = mol.atoms[ai - 1]
                            pair.atom2 = mol.atoms[aj - 1]
                            pair.gromacs["func"] = fu

                            mol.pairs.append(pair)
                            _add_info(mol, curr_sec, mol.pairs)

                        else:
                            raise ValueError

                    else:
                        raise NotImplementedError(
                            "{0:s} with functiontype {1:d} is not supported".format(
                                curr_sec, fu
                            )
                        )

                elif curr_sec == "nonbond_params":
                    """
                    ; typei typej  f.type sigma   epsilon
                    ; f.type=1 means LJ (not buckingham)
                    ; sigma&eps since mixing-rule = 2
                    """

                    assert len(fields) == 5
                    ai, aj = fields[:2]
                    fu = int(fields[2])

                    assert fu == 1
                    sig = float(fields[3])
                    eps = float(fields[4])

                    nonbond_param = blocks.NonbondedParamType("gromacs")
                    nonbond_param.atype1 = ai
                    nonbond_param.atype2 = aj
                    nonbond_param.gromacs["func"] = fu
                    nonbond_param.gromacs["param"] = {"eps": eps, "sig": sig}

                    self.nonbond_params.append(nonbond_param)
                    _add_info(self, curr_sec, self.nonbond_params)

                elif curr_sec in ("bondtypes", "bonds"):
                    """
                    section     #at     fu      #param
                    ----------------------------------
                    bonds       2       1       2
                    bonds       2       2       2
                    bonds       2       3       3
                    bonds       2       4       2
                    bonds       2       5       ??
                    bonds       2       6       2
                    bonds       2       7       2
                    bonds       2       8       ??
                    bonds       2       9       ??
                    bonds       2       10      4
                    """

                    ai, aj = fields[:2]
                    fu = int(fields[2])
                    assert fu in (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

                    if fu != 1:
                        raise NotImplementedError(
                            "function {0:d} is not yet supported".format(fu)
                        )

                    bond = blocks.BondType("gromacs")

                    if fu == 1:
                        if curr_sec == "bondtypes":
                            bond.atype1 = ai
                            bond.atype2 = aj

                            b0, kb = list(map(float, fields[3:5]))
                            bond.gromacs = {"param": {"kb": kb, "b0": b0}, "func": fu}

                            self.bondtypes.append(bond)
                            _add_info(self, curr_sec, self.bondtypes)

                        elif curr_sec == "bonds":
                            ai, aj = list(map(int, [ai, aj]))
                            bond.atom1 = mol.atoms[ai - 1]
                            bond.atom2 = mol.atoms[aj - 1]
                            bond.gromacs["func"] = fu

                            if len(fields) > 3:
                                b0, kb = list(map(float, fields[3:5]))
                                bond.gromacs = {
                                    "param": {"kb": kb, "b0": b0},
                                    "func": fu,
                                }

                            mol.bonds.append(bond)
                            _add_info(mol, curr_sec, mol.bonds)

                    else:
                        raise NotImplementedError

                elif curr_sec in ("angletypes", "angles"):
                    """
                    section     #at     fu      #param
                    ----------------------------------
                    angles      3       1       2
                    angles      3       2       2
                    angles      3       3       3
                    angles      3       4       4
                    angles      3       5       4
                    angles      3       6       6
                    angles      3       8       ??
                    """

                    ai, aj, ak = fields[:3]
                    fu = int(fields[3])
                    assert fu in (1, 2, 3, 4, 5, 6, 8)  # no 7

                    if fu not in (1, 2, 5):
                        raise NotImplementedError(
                            "function {0:d} is not yet supported".format(fu)
                        )

                    ang = blocks.AngleType("gromacs")
                    if fu == 1:
                        if curr_sec == "angletypes":
                            ang.atype1 = ai
                            ang.atype2 = aj
                            ang.atype3 = ak

                            tetha0, ktetha = list(map(float, fields[4:6]))
                            ang.gromacs = {
                                "param": {
                                    "ktetha": ktetha,
                                    "tetha0": tetha0,
                                    "kub": None,
                                    "s0": None,
                                },
                                "func": fu,
                            }

                            self.angletypes.append(ang)
                            _add_info(self, curr_sec, self.angletypes)

                        elif curr_sec == "angles":
                            ai, aj, ak = list(map(int, [ai, aj, ak]))
                            ang.atom1 = mol.atoms[ai - 1]
                            ang.atom2 = mol.atoms[aj - 1]
                            ang.atom3 = mol.atoms[ak - 1]
                            ang.gromacs["func"] = fu

                            mol.angles.append(ang)
                            _add_info(mol, curr_sec, mol.angles)

                        else:
                            raise ValueError

                    elif fu == 2:
                        if curr_sec == "angletypes":
                            raise NotImplementedError()

                        elif curr_sec == "angles":
                            ai, aj, ak = list(map(int, [ai, aj, ak]))
                            ang.atom1 = mol.atoms[ai - 1]
                            ang.atom2 = mol.atoms[aj - 1]
                            ang.atom3 = mol.atoms[ak - 1]
                            ang.gromacs["func"] = fu

                            tetha0, ktetha = list(map(float, fields[4:6]))
                            ang.gromacs = {
                                "param": {
                                    "ktetha": ktetha,
                                    "tetha0": tetha0,
                                    "kub": None,
                                    "s0": None,
                                },
                                "func": fu,
                            }

                            mol.angles.append(ang)
                            _add_info(mol, curr_sec, mol.angles)

                    elif fu == 5:
                        if curr_sec == "angletypes":
                            ang.atype1 = ai
                            ang.atype2 = aj
                            ang.atype3 = ak
                            tetha0, ktetha, s0, kub = list(map(float, fields[4:8]))

                            ang.gromacs = {
                                "param": {
                                    "ktetha": ktetha,
                                    "tetha0": tetha0,
                                    "kub": kub,
                                    "s0": s0,
                                },
                                "func": fu,
                            }

                            self.angletypes.append(ang)
                            _add_info(self, curr_sec, self.angletypes)

                        elif curr_sec == "angles":
                            ai, aj, ak = list(map(int, [ai, aj, ak]))
                            ang.atom1 = mol.atoms[ai - 1]
                            ang.atom2 = mol.atoms[aj - 1]
                            ang.atom3 = mol.atoms[ak - 1]
                            ang.gromacs["func"] = fu

                            mol.angles.append(ang)
                            _add_info(mol, curr_sec, mol.angles)

                        else:
                            raise ValueError

                    else:
                        raise NotImplementedError

                elif curr_sec in ("dihedraltypes", "dihedrals"):
                    """
                    section     #at     fu      #param
                    ----------------------------------
                    dihedrals   4       1       3
                    dihedrals   4       2       2
                    dihedrals   4       3       6
                    dihedrals   4       4       3
                    dihedrals   4       5       4
                    dihedrals   4       8       ??
                    dihedrals   4       9       3
                    """

                    if curr_sec == "dihedraltypes" and len(fields) == 6:
                        # in oplsaa - quartz parameters
                        fields.insert(2, "X")
                        fields.insert(0, "X")

                    ai, aj, ak, am = fields[:4]
                    fu = int(fields[4])
                    assert fu in (1, 2, 3, 4, 5, 8, 9)

                    if fu not in (1, 2, 3, 4, 9):
                        raise NotImplementedError(
                            "dihedral function {0:d} is not yet supported".format(fu)
                        )

                    dih = blocks.DihedralType("gromacs")
                    imp = blocks.ImproperType("gromacs")
                    # proper dihedrals
                    if fu in (1, 3, 9):
                        if curr_sec == "dihedraltypes":
                            dih.atype1 = ai
                            dih.atype2 = aj
                            dih.atype3 = ak
                            dih.atype4 = am

                            dih.line = i_line + 1

                            if fu == 1:
                                delta, kchi, n = list(map(float, fields[5:8]))
                                dih.gromacs["param"].append(
                                    {"kchi": kchi, "n": n, "delta": delta}
                                )
                            elif fu == 3:
                                c0, c1, c2, c3, c4, c5 = list(map(float, fields[5:11]))
                                m = dict(c0=c0, c1=c1, c2=c2, c3=c3, c4=c4, c5=c5)
                                dih.gromacs["param"].append(m)
                            elif fu == 4:
                                delta, kchi, n = list(map(float, fields[5:8]))
                                dih.gromacs["param"].append(
                                    {"kchi": kchi, "n": int(n), "delta": delta}
                                )
                            elif fu == 9:
                                delta, kchi, n = list(map(float, fields[5:8]))
                                dih.gromacs["param"].append(
                                    {"kchi": kchi, "n": int(n), "delta": delta}
                                )
                            else:
                                raise ValueError

                            dih.gromacs["func"] = fu
                            self.dihedraltypes.append(dih)
                            _add_info(self, curr_sec, self.dihedraltypes)

                        elif curr_sec == "dihedrals":
                            ai, aj, ak, am = list(map(int, fields[:4]))
                            dih.atom1 = mol.atoms[ai - 1]
                            dih.atom2 = mol.atoms[aj - 1]
                            dih.atom3 = mol.atoms[ak - 1]
                            dih.atom4 = mol.atoms[am - 1]
                            dih.gromacs["func"] = fu

                            dih.line = i_line + 1

                            if fu == 1:
                                delta, kchi, n = list(map(float, fields[5:8]))
                                dih.gromacs["param"].append(
                                    {"kchi": kchi, "n": int(n), "delta": delta}
                                )
                            elif fu == 3:
                                pass
                            elif fu == 4:
                                pass
                            elif fu == 9:
                                if len(fields[5:8]) == 3:
                                    delta, kchi, n = list(map(float, fields[5:8]))
                                    dih.gromacs["param"].append(
                                        {"kchi": kchi, "n": int(n), "delta": delta}
                                    )
                            else:
                                raise ValueError

                            mol.dihedrals.append(dih)
                            _add_info(mol, curr_sec, mol.dihedrals)

                        else:
                            raise ValueError
                    # impropers
                    elif fu in (2, 4):
                        if curr_sec == "dihedraltypes":
                            imp.atype1 = ai
                            imp.atype2 = aj
                            imp.atype3 = ak
                            imp.atype4 = am

                            imp.line = i_line + 1

                            if fu == 2:
                                psi0, kpsi = list(map(float, fields[5:7]))
                                imp.gromacs["param"].append(
                                    {"kpsi": kpsi, "psi0": psi0}
                                )
                            elif fu == 4:
                                psi0, kpsi, n = list(map(float, fields[5:8]))
                                imp.gromacs["param"].append(
                                    {"kpsi": kpsi, "psi0": psi0, "n": int(n)}
                                )
                            else:
                                raise ValueError

                            imp.gromacs["func"] = fu
                            self.impropertypes.append(imp)
                            _add_info(self, curr_sec, self.impropertypes)

                        elif curr_sec == "dihedrals":
                            ai, aj, ak, am = list(map(int, fields[:4]))
                            imp.atom1 = mol.atoms[ai - 1]
                            imp.atom2 = mol.atoms[aj - 1]
                            imp.atom3 = mol.atoms[ak - 1]
                            imp.atom4 = mol.atoms[am - 1]
                            imp.gromacs["func"] = fu

                            imp.line = i_line + 1

                            if fu == 2:
                                pass
                            elif fu == 4:
                                # in-line override of dihedral parameters
                                if len(fields[5:8]) == 3:
                                    psi0, kpsi, n = list(map(float, fields[5:8]))
                                    imp.gromacs["param"].append(
                                        {"kpsi": kpsi, "psi0": psi0, "n": int(n)}
                                    )
                            else:
                                raise ValueError

                            mol.impropers.append(imp)
                            _add_info(mol, curr_sec, mol.impropers)

                        else:
                            raise ValueError

                    else:
                        raise NotImplementedError

                elif curr_sec in ("cmaptypes", "cmap"):
                    cmap = blocks.CMapType("gromacs")
                    if curr_sec == "cmaptypes":
                        cmap_lines.append(line)
                        _add_info(self, curr_sec, self.cmaptypes)
                    else:
                        ai, aj, ak, am, an = list(map(int, fields[:5]))
                        fu = int(fields[5])
                        assert fu == 1
                        cmap.atom1 = mol.atoms[ai - 1]
                        cmap.atom2 = mol.atoms[aj - 1]
                        cmap.atom3 = mol.atoms[ak - 1]
                        cmap.atom4 = mol.atoms[am - 1]
                        cmap.atom8 = mol.atoms[an - 1]
                        cmap.gromacs["func"] = fu

                        mol.cmaps.append(cmap)
                        _add_info(mol, curr_sec, mol.cmaps)

                elif curr_sec == "settles":
                    """
                    section     #at     fu      #param
                    ----------------------------------
                    """

                    assert len(fields) == 4
                    ai = int(fields[0])
                    fu = int(fields[1])
                    assert fu == 1

                    settle = blocks.SettleType("gromacs")
                    settle.atom = mol.atoms[ai - 1]
                    settle.dOH = float(fields[2])
                    settle.dHH = float(fields[3])

                    mol.settles.append(settle)
                    _add_info(mol, curr_sec, mol.settles)

                elif curr_sec == "virtual_sites3":
                    """
                    ; Dummy from            funct   a       b
                    4   1   2   3   1   0.131937768 0.131937768
                    """
                    assert len(fields) == 7
                    ai = int(fields[0])
                    aj = int(fields[1])
                    ak = int(fields[2])
                    al = int(fields[3])
                    fu = int(fields[4])
                    assert fu == 1
                    a = float(fields[5])
                    b = float(fields[6])

                    vs3 = blocks.VirtualSites3Type("gromacs")
                    vs3.atom1 = ai
                    vs3.atom2 = aj
                    vs3.atom3 = ak
                    vs3.atom4 = al
                    vs3.gromacs["func"] = fu
                    vs3.gromacs["param"] = {"a": a, "b": b}
                    mol.virtual_sites3.append(vs3)
                    _add_info(mol, curr_sec, mol.virtual_sites3)

                elif curr_sec in ("exclusions",):
                    ai = int(fields[0])
                    other = list(map(int, fields[1:]))

                    exc = blocks.Exclusion()
                    exc.main_atom = mol.atoms[ai - 1]
                    exc.other_atoms = [mol.atoms[k - 1] for k in other]

                    mol.exclusions.append(exc)
                    _add_info(mol, curr_sec, mol.exclusions)

                elif curr_sec in ("constrainttypes", "constraints"):
                    """
                    section     #at     fu      #param
                    ----------------------------------
                    constraints 2       1       1
                    constraints 2       2       1
                    """

                    ai, aj = fields[:2]
                    fu = int(fields[2])
                    assert fu in (1, 2)

                    cons = blocks.ConstraintType("gromacs")

                    # TODO: what's different between 1 and 2
                    if fu in [1, 2]:
                        if curr_sec == "constrainttypes":
                            cons.atype1 = ai
                            cons.atype2 = aj
                            b0 = float(fields[3])
                            cons.gromacs = {"param": {"b0": b0}, "func": fu}

                            self.constrainttypes.append(cons)
                            _add_info(self, curr_sec, self.constrainttypes)

                        elif curr_sec == "constraints":
                            ai, aj = list(map(int, fields[:2]))
                            cons.atom1 = mol.atoms[ai - 1]
                            cons.atom2 = mol.atoms[aj - 1]
                            cons.gromacs["func"] = fu

                            mol.constraints.append(cons)
                            _add_info(mol, curr_sec, mol.constraints)

                        else:
                            raise ValueError
                    else:
                        raise ValueError

                elif curr_sec in (
                    "position_restraints",
                    "distance_restraints",
                    "dihedral_restraints",
                    "orientation_restraints",
                    "angle_restraints",
                    "angle_restraints_z",
                ):
                    pass

                elif curr_sec in ("implicit_genborn_params",):
                    """
                    attype   sar     st      pi      gbr      hct
                    """
                    pass

                elif curr_sec == "system":
                    # assert len(fields) == 1
                    self.name = fields[0]

                elif curr_sec == "molecules":
                    assert len(fields) == 2
                    mname, nmol = fields[0], int(fields[1])

                    # if the number of a molecule is more than 1, add copies to system.molecules
                    for i in range(nmol):
                        self.molecules.append(self.dict_molname_mol[mname])

                else:
                    raise NotImplementedError(
                        "Unknown section in topology: {0}".format(curr_sec)
                    )

        # process cmap_lines
        curr_cons = None
        for line in cmap_lines:
            # cmaptype opening line
            if len(line.split()) == 8:
                cons = blocks.CMapType("gromacs")

                (
                    atype1,
                    atype2,
                    atype3,
                    atype4,
                    atype8,
                    func,
                    sizeX,
                    sizeY,
                ) = line.replace("\\", "").split()
                func, sizeX, sizeY = int(func), int(sizeX), int(sizeY)
                cons.atype1 = atype1
                cons.atype2 = atype2
                cons.atype3 = atype3
                cons.atype4 = atype4
                cons.atype8 = atype8
                cons.gromacs = {"param": [], "func": func}

                curr_cons = cons

            # cmap body
            elif len(line.split()) == 10:
                cmap_param = map(float, line.replace("\\", "").split())
                cons.gromacs["param"] += cmap_param

            # cmaptype cloning line
            elif len(line.split()) == 6:
                cmap_param = map(float, line.replace("\\", "").split())
                cons.gromacs["param"] += cmap_param
                self.cmaptypes.append(curr_cons)
            else:
                raise ValueError


class SystemToGroTop(object):
    """Converter class - represent TOP objects as GROMACS topology file."""

    formats = {
        "atomtypes": "{:<7s} {:3s} {:>7} {} {:3s} {} {}\n",
        "atoms": "{:6d} {:>10s} {:6d} {:6s} {:6s} {:6d} {} {}\n",
        "atoms_nomass": "{:6d} {:>10s} {:6d} {:6s} {:6s} {:6d} {}\n",
        "nonbond_params": "{:20s}  {:20s}  {:1d}  {}  {}\n",
        "bondtypes": "{:5s}  {:5s}  {:1d}  {}  {}\n",
        "bonds": "{:3d}  {:3d}   {:1d}\n",
        "bonds_ext": "{:3d}  {:3d}   {:1d} {} {}\n",
        "settles": "{:3d}  {:3d}  {} {}\n",
        "virtual_sites3": "{:3d}  {:3d}  {:3d}  {:3d}   {:1d}  {}  {}\n",
        "exclusions": "{:3d}  {}\n",
        "pairtypes": "{:6s} {:6s}   {:d}    {:.13g}     {:.13g}\n",
        "pairs": "{:3d} {:3d}   {:1d}\n",
        "angletypes_1": "{:>8s} {:>8s} {:>8s} {:1d}    {}    {}\n",
        "angletypes_5": "{:>8s} {:>8s} {:>8s} {:1d}    {}    {}    {}    {}\n",
        "constrainttypes": "{:6s} {:6s} {:1d}    {}\n",
        "angles": "{:3d} {:3d} {:3d}   {:1d}\n",
        "angles_ext": "{:3d} {:3d} {:3d}   {:1d} {} {}\n",
        "dihedraltypes": "{:6s} {:6s} {:6s} {:6s}   {:1d}    {}    {}    {:1d}\n",
        "dihedrals": "{:3d} {:3d} {:3d} {:3d}   {:1d}\n",
        "dihedrals_ext": "{:3d} {:3d} {:3d} {:3d}   {:1d}    {}    {}    {:1d}\n",
        "impropertypes_2": "{:6s} {:6s} {:6s} {:6s}   {:1d} {} {} \n",
        "impropertypes_4": "{:6s} {:6s} {:6s} {:6s}   {:1d} {} {} {:2d}\n",
        "impropers": "{:3d} {:3d} {:3d} {:3d}   {:1d}\n",
        "impropers_2": "{:3d} {:3d} {:3d} {:3d}   {:1d} {} {} \n",
        "impropers_4": "{:3d} {:3d} {:3d} {:3d}   {:1d} {} {} {:2d}\n",
    }

    toptemplate = """
            [ defaults ]
            *DEFAULTS*
            [ atomtypes ]
            *ATOMTYPES*
            [ nonbond_params ]
            *NONBOND_PARAM*
            [ pairtypes ]
            *PAIRTYPES*
            [ bondtypes ]
            *BONDTYPES*
            [ angletypes ]
            *ANGLETYPES*
            [ constrainttypes ]
            *CONSTRAINTTYPES*
            [ dihedraltypes ]
            *DIHEDRALTYPES*
            [ dihedraltypes ]
            *IMPROPERTYPES*
            [ cmaptypes ]
            *CMAPTYPES*
            """
    toptemplate = textwrap.dedent(toptemplate)

    itptemplate = """
            [ moleculetype ]
            *MOLECULETYPE*
            [ atoms ]
            *ATOMS*
            [ bonds ]
            *BONDS*
            [ pairs ]
            *PAIRS*
            [ settles ]
            *SETTLES*
            [ virtual_sites3 ]
            *VIRTUAL_SITES3*
            [ exclusions ]
            *EXCLUSIONS*
            [ angles ]
            *ANGLES*
            [ dihedrals ]
            *DIHEDRALS*
            [ dihedrals ]
            *IMPROPERS*
            [ cmap ]
            *CMAPS*
            """
    itptemplate = textwrap.dedent(itptemplate)

    def __init__(self, system, outfile="output.top", multiple_output=False):
        """Initialize GROMACS topology writer.

        :Arguments:
          *system*
              :class:`blocks.System` object, containing the topology
          *outfile*
              name of the file to write to
          *multiple_output*
              if True, write moleculetypes to separate files, named mol_MOLNAME.itp (default: False)
        """
        self.logger = logging.getLogger("gromacs.fileformats.SystemToGroTop")
        self.logger.debug(">> entering SystemToGroTop")

        self.system = system
        self.outfile = outfile
        self.multiple_output = multiple_output
        self.assemble_topology()

        self.logger.debug("<< leaving SystemToGroTop")

    @staticmethod
    def _redefine_atomtypes(mol):
        for i, atom in enumerate(mol.atoms):
            atom.atomtype = "at{0:03d}".format(i + 1)

    def assemble_topology(self):
        """Call the various member self._make_* functions to convert the topology object into a string"""
        self.logger.debug("starting to assemble topology...")

        top = ""

        self.logger.debug("making atom/pair/bond/angle/dihedral/improper types")
        top += self.toptemplate
        top = top.replace("*DEFAULTS*", "".join(self._make_defaults(self.system)))
        top = top.replace("*ATOMTYPES*", "".join(self._make_atomtypes(self.system)))
        top = top.replace(
            "*NONBOND_PARAM*", "".join(self._make_nonbond_param(self.system))
        )
        top = top.replace("*PAIRTYPES*", "".join(self._make_pairtypes(self.system)))
        top = top.replace("*BONDTYPES*", "".join(self._make_bondtypes(self.system)))
        top = top.replace(
            "*CONSTRAINTTYPES*", "".join(self._make_constrainttypes(self.system))
        )
        top = top.replace("*ANGLETYPES*", "".join(self._make_angletypes(self.system)))
        top = top.replace(
            "*DIHEDRALTYPES*", "".join(self._make_dihedraltypes(self.system))
        )
        top = top.replace(
            "*IMPROPERTYPES*", "".join(self._make_impropertypes(self.system))
        )
        top = top.replace("*CMAPTYPES*", "".join(self._make_cmaptypes(self.system)))

        for i, (molname, m) in enumerate(self.system.dict_molname_mol.items()):
            itp = self.itptemplate
            itp = itp.replace(
                "*MOLECULETYPE*",
                "".join(self._make_moleculetype(m, molname, m.exclusion_numb)),
            )
            itp = itp.replace("*ATOMS*", "".join(self._make_atoms(m)))
            itp = itp.replace("*BONDS*", "".join(self._make_bonds(m)))
            itp = itp.replace("*PAIRS*", "".join(self._make_pairs(m)))
            itp = itp.replace("*SETTLES*", "".join(self._make_settles(m)))
            itp = itp.replace("*VIRTUAL_SITES3*", "".join(self._make_virtual_sites3(m)))
            itp = itp.replace("*EXCLUSIONS*", "".join(self._make_exclusions(m)))
            itp = itp.replace("*ANGLES*", "".join(self._make_angles(m)))
            itp = itp.replace("*DIHEDRALS*", "".join(self._make_dihedrals(m)))
            itp = itp.replace("*IMPROPERS*", "".join(self._make_impropers(m)))
            itp = itp.replace("*CMAPS*", "".join(self._make_cmaps(m)))
            if not self.multiple_output:
                top += itp
            else:
                outfile = "mol_{0}.itp".format(molname)
                top += '#include "mol_{0}.itp" \n'.format(molname)
                with open(outfile, "w") as f:
                    f.writelines([itp])

        top += "\n[system]  \nConvertedSystem\n\n"
        top += "[molecules] \n"
        molecules = [("", 0)]

        for m in self.system.molecules:
            if molecules[-1][0] != m.name:
                molecules.append([m.name, 0])
            if molecules[-1][0] == m.name:
                molecules[-1][1] += 1

        for molname, n in molecules[1:]:
            top += "{0:s}     {1:d}\n".format(molname, n)
        top += "\n"

        with open(self.outfile, "w") as f:
            f.writelines([top])

    def _make_defaults(self, m):
        if m.defaults["gen-pairs"] and m.defaults["fudgeLJ"] and m.defaults["fudgeQQ"]:
            line = [
                "{0:d}          {1:d}           {2}          {3}       {4} \n".format(
                    m.defaults["nbfunc"],
                    m.defaults["comb-rule"],
                    m.defaults["gen-pairs"],
                    m.defaults["fudgeLJ"],
                    m.defaults["fudgeQQ"],
                )
            ]
        else:
            line = [
                "{0:d}          {1:d}\n".format(
                    m.defaults["nbfunc"],
                    m.defaults["comb-rule"],
                )
            ]
        return line

    def _make_atomtypes(self, m):
        def get_prot(at):
            # TODO improve this
            _protons = {"C": 6, "H": 1, "N": 7, "O": 8, "S": 16, "P": 15}
            if at[0] in list(_protons.keys()):
                return _protons[at[0]]
            else:
                return 0

        result = []
        for at in m.atomtypes:
            at.convert("gromacs")
            prot = get_prot(at.atype)
            ljl = at.gromacs["param"]["ljl"]
            lje = at.gromacs["param"]["lje"]
            line = self.formats["atomtypes"].format(
                at.atype,
                at.bond_type if at.bond_type else "",
                at.mass,
                at.charge,
                "A",
                ljl,
                lje,
            )
            if at.comment:
                line += at.comment
            result.append(line)

        return result

    def _make_nonbond_param(self, m):
        result = []
        for pr in m.nonbond_params:
            at1 = pr.atype1
            at2 = pr.atype2

            # pr.convert('gromacs')
            eps = pr.gromacs["param"]["eps"]
            sig = pr.gromacs["param"]["sig"]

            fu = 1  # TODO
            line = self.formats["nonbond_params"].format(at1, at2, fu, sig, eps)
            if pr.comment:
                line = line[:-1] + pr.comment + line[-1:]
            result.append(line)

        return result

    def _make_pairtypes(self, m):
        result = []
        for pt in m.pairtypes:
            at1, at2 = pt.atype1, pt.atype2
            fu, l14, e14 = (
                pt.gromacs["func"],
                pt.gromacs["param"]["ljl14"],
                pt.gromacs["param"]["lje14"],
            )
            line = self.formats["pairtypes"].format(at1, at2, fu, l14, e14)
            if pt.comment:
                line = line[:-1] + pt.comment
            result.append(line)

        return result

    def _make_bondtypes(self, m):
        result = []
        for bond in m.bondtypes:
            at1 = bond.atype1
            at2 = bond.atype2
            bond.convert("gromacs")

            kb = bond.gromacs["param"]["kb"]
            b0 = bond.gromacs["param"]["b0"]
            fu = bond.gromacs["func"]

            line = self.formats["bondtypes"].format(at1, at2, fu, b0, kb)
            result.append(line)

        return result

    def _make_constrainttypes(self, m):
        result = []

        for con in m.constrainttypes:
            at1 = con.atype1
            at2 = con.atype2

            fu = con.gromacs["func"]
            b0 = con.gromacs["param"]["b0"]

            line = self.formats["constrainttypes"].format(at1, at2, fu, b0)
            result.append(line)

        return result

    def _make_angletypes(self, m):
        result = []
        for ang in m.angletypes:
            at1 = ang.atype1
            at2 = ang.atype2
            at3 = ang.atype3
            ang.convert("gromacs")

            ktetha = ang.gromacs["param"]["ktetha"]
            tetha0 = ang.gromacs["param"]["tetha0"]
            kub = ang.gromacs["param"]["kub"]
            s0 = ang.gromacs["param"]["s0"]

            fu = ang.gromacs["func"]

            angletypes = "angletypes_{0:d}".format(fu)
            line = self.formats[angletypes].format(
                at1, at2, at3, fu, tetha0, ktetha, s0, kub
            )
            result.append(line)

        return result

    def _make_dihedraltypes(self, m):
        result = []
        for dih in m.dihedraltypes:
            at1 = dih.atype1
            at2 = dih.atype2
            at3 = dih.atype3
            at4 = dih.atype4

            dih.convert("gromacs")
            fu = dih.gromacs["func"]

            for dpar in dih.gromacs["param"]:
                kchi = dpar["kchi"]
                n = dpar["n"]
                delta = dpar["delta"]

                if not dih.disabled:
                    line = self.formats["dihedraltypes"].format(
                        at1, at2, at3, at4, fu, delta, kchi, n
                    )
                else:
                    line = self.formats["dihedraltypes"].format(
                        at1, at2, at3, at4, fu, delta, kchi, n
                    )
                    line = dih.comment + line
                result.append(line)

        return result

    def _make_impropertypes(self, m):
        result = []
        for imp in m.impropertypes:
            at1 = imp.atype1
            at2 = imp.atype2
            at3 = imp.atype3
            at4 = imp.atype4

            imp.convert("gromacs")
            fu = imp.gromacs["func"]

            for ipar in imp.gromacs["param"]:
                kpsi = ipar["kpsi"]
                psi0 = ipar["psi0"]

                if fu == 2:
                    line = self.formats["impropertypes_2"].format(
                        at1, at2, at3, at4, fu, psi0, kpsi
                    )
                if fu == 4:
                    n = ipar["n"]
                    line = self.formats["impropertypes_4"].format(
                        at1, at2, at3, at4, fu, psi0, kpsi, n
                    )

                if imp.disabled:
                    line = imp.comment + line
                result.append(line)

        return result

    def _make_cmaptypes(self, m):
        result = []
        for cmap in m.cmaptypes:
            at1 = cmap.atype1
            at2 = cmap.atype2
            at3 = cmap.atype3
            at4 = cmap.atype4
            # at5 = cmap.atype5
            # at6 = cmap.atype6
            # at7 = cmap.atype7
            at8 = cmap.atype8

            cmap.convert("gromacs")

            fu = cmap.gromacs["func"]
            line = "{0:s} {1:s} {2:s} {3:s} {4:s} {5:d} 24 24".format(
                at1, at2, at3, at4, at8, fu
            )
            for i, c in enumerate(cmap.gromacs["param"]):
                if i % 10 == 0:
                    line += "\\\n"
                else:
                    line += " "
                line += "{0:12.8f}".format(c)

            line += "\n\n"
            result.append(line)

        return result

    def _make_moleculetype(self, m, molname, nrexcl):
        return ["; Name \t\t  nrexcl \n {0}    {1} \n".format(molname, nrexcl)]

    def _make_atoms(self, m):
        result = []
        # i = 1
        for atom in m.atoms:
            numb = cgnr = atom.number
            atype = atom.get_atomtype()

            assert atype != False
            assert hasattr(atom, "charge")  # and hasattr(atom, 'mass')

            if hasattr(atom, "mass"):
                line = self.formats["atoms"].format(
                    numb,
                    atype,
                    atom.resnumb,
                    atom.resname,
                    atom.name,
                    cgnr,
                    atom.charge,
                    atom.mass,
                )
            else:
                line = self.formats["atoms_nomass"].format(
                    numb,
                    atype,
                    atom.resnumb,
                    atom.resname,
                    atom.name,
                    cgnr,
                    atom.charge,
                )
            result.append(line)

        result.insert(0, "; {0:5d} atoms\n".format(len(result)))
        return result

    def _make_pairs(self, m):
        result = []
        for pr in m.pairs:
            fu = 1
            p1 = pr.atom1.number
            p4 = pr.atom2.number

            line = self.formats["pairs"].format(p1, p4, fu)
            result.append(line)

        result.insert(0, "; {0:5d} pairs\n".format(len(result)))
        return result

    def _make_bonds(self, m):
        result = []
        for bond in m.bonds:
            fu = bond.gromacs["func"]

            if bond.gromacs["param"]["kb"] and bond.gromacs["param"]["b0"]:
                kb, b0 = bond.gromacs["param"]["kb"], bond.gromacs["param"]["b0"]
                line = self.formats["bonds_ext"].format(
                    bond.atom1.number, bond.atom2.number, fu, b0, kb
                )
            else:
                line = self.formats["bonds"].format(
                    bond.atom1.number, bond.atom2.number, fu
                )

            result.append(line)

        result.insert(0, "; {0:5d} bonds\n".format(len(result)))
        return result

    def _make_angles(self, m):
        result = []
        for ang in m.angles:
            fu = ang.gromacs["func"]
            if ang.gromacs["param"]["ktetha"] and ang.gromacs["param"]["tetha0"]:
                ktetha, tetha0 = (
                    ang.gromacs["param"]["ktetha"],
                    ang.gromacs["param"]["tetha0"],
                )
                line = self.formats["angles_ext"].format(
                    ang.atom1.number,
                    ang.atom2.number,
                    ang.atom3.number,
                    fu,
                    tetha0,
                    ktetha,
                )
            else:
                line = self.formats["angles"].format(
                    ang.atom1.number, ang.atom2.number, ang.atom3.number, fu
                )
            result.append(line)

        result.insert(0, "; {0:5d} angles\n".format(len(result)))
        return result

    def _make_settles(self, m):
        result = []
        for st in m.settles:
            line = self.formats["settles"].format(st.atom.number, 1, st.dOH, st.dHH)
            result.append(line)

        result.insert(0, "; {0:5d} settles\n".format(len(result)))
        return result

    def _make_virtual_sites3(self, m):
        result = []
        for vs in m.virtual_sites3:
            fu = 1
            line = self.formats["virtual_sites3"].format(
                vs.atom1,
                vs.atom2,
                vs.atom3,
                vs.atom4,
                fu,
                vs.gromacs["param"]["a"],
                vs.gromacs["param"]["b"],
            )
            result.append(line)

        result.insert(0, "; {0:5d} virtual_sites3\n".format(len(result)))
        return result

    def _make_exclusions(self, m):
        result = []
        for excl in m.exclusions:
            other_atoms = ["  {:3d}".format(at.number) for at in excl.other_atoms]
            line = self.formats["exclusions"].format(
                excl.main_atom.number, "".join(other_atoms)
            )
            result.append(line)

        result.insert(0, "; {0:5d} exclusions\n".format(len(result)))
        return result

    def _make_dihedrals(self, m):
        result = []
        for dih in m.dihedrals:
            fu = dih.gromacs["func"]

            if not dih.gromacs["param"]:
                line = self.formats["dihedrals"].format(
                    dih.atom1.number,
                    dih.atom2.number,
                    dih.atom3.number,
                    dih.atom4.number,
                    fu,
                )
                result.append(line)

            for dpar in dih.gromacs["param"]:
                kchi = dpar["kchi"]
                n = dpar["n"]
                delta = dpar["delta"]

                line = self.formats["dihedrals_ext"].format(
                    dih.atom1.number,
                    dih.atom2.number,
                    dih.atom3.number,
                    dih.atom4.number,
                    fu,
                    delta,
                    kchi,
                    n,
                )
                if dih.comment:
                    line = dih.comment + line
                result.append(line)

        result.insert(0, "; {0:5d} dihedrals\n".format(len(result)))
        return result

    def _make_impropers(self, m):
        result = []
        for imp in m.impropers:
            fu = imp.gromacs["func"]

            if not imp.gromacs["param"]:
                line = self.formats["impropers"].format(
                    imp.atom1.number,
                    imp.atom2.number,
                    imp.atom3.number,
                    imp.atom4.number,
                    fu,
                )
                result.append(line)

            for ipar in imp.gromacs["param"]:
                kpsi = ipar["kpsi"]
                psi0 = ipar["psi0"]

                if fu == 2:
                    line = self.formats["impropers_2"].format(
                        imp.atom1.number,
                        imp.atom2.number,
                        imp.atom3.number,
                        imp.atom4.number,
                        fu,
                        psi0,
                        kpsi,
                    )
                if fu == 4:
                    n = ipar["n"]
                    line = self.formats["impropers_4"].format(
                        imp.atom1.number,
                        imp.atom2.number,
                        imp.atom3.number,
                        imp.atom4.number,
                        fu,
                        psi0,
                        kpsi,
                        n,
                    )

                if imp.comment:
                    line = imp.comment + line
                result.append(line)

        result.insert(0, "; {0:5d} impropers\n".format(len(result)))
        return result

    def _make_cmaps(self, m):
        result = []

        for cmap in m.cmaps:
            fu = 1
            line = "{0:5d} {1:5d} {2:5d} {3:5d} {4:5d}   {5:d}\n".format(
                cmap.atom1.number,
                cmap.atom2.number,
                cmap.atom3.number,
                cmap.atom4.number,
                cmap.atom8.number,
                fu,
            )
            result.append(line)

        result.insert(0, "; {0:5d} cmaps\n".format(len(result)))
        return result



# -*- coding: utf-8 -*-
# GromacsWrapper
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import six

import pytest

import gromacs


@pytest.fixture(
    scope="function",
    params=[
        [0, "foo", None, 2.7e-2, "foo"],
        (0, "foo", None, 2.7e-2, "foo"),
        set([0, "foo", None, 2.7e-2]),
        ["ant", "boar", "ape", "gnu"],
        [["ant", "spider"], ["boar", "ape", "gnu"]],
    ],
)
def things(request):
    stuff = request.param
    return stuff, gromacs.collections.Collection(stuff)


@pytest.fixture(
    scope="function",
    params=[["ant", "boar", "ape", "gnu"], ["åmeise", "Beißfliege", "Ürmelchen"]],
)
def textthings(request):
    stuff = request.param
    return stuff, gromacs.collections.Collection(stuff)


class TestCollection(object):
    def test_list_like(self, things):
        seq, collection = things
        assert isinstance(collection, list)
        assert len(collection) == len(seq)

    def test_tolist(self, things):
        seq, collection = things
        lst = collection.tolist()
        assert lst == list(seq)

    def test_save_load(self, things, tmpdir):
        fname = "stuff"
        seq, collection = things
        with tmpdir.as_cwd():
            collection.save(fname)
            newcollection = gromacs.collections.Collection()
            newcollection.load(fname)
        assert newcollection.tolist() == list(seq)
        assert newcollection == collection

    @pytest.mark.parametrize(
        "method,args", [("startswith", ("å",)), ("upper", ()), ("capitalize", ())]
    )
    def test_method_pass_through(self, textthings, method, args):
        seq, collection = textthings
        results = getattr(collection, method)(*args)
        assert results == [getattr(elem, method)(*args) for elem in seq]

    @pytest.mark.parametrize("attribute", ["__doc__"])
    def test_attribute_pass_through(self, textthings, attribute):
        _, collection = textthings
        results = getattr(collection, attribute)
        assert results

    def test_add(self, textthings):
        seq, collection = textthings
        bigger = collection + collection
        assert isinstance(bigger, gromacs.collections.Collection)

        seq.extend(seq)
        assert bigger.tolist() == list(seq)



# GromacsWrapper: test_example.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import os.path
import pytest

from numpy.testing import assert_almost_equal

import gromacs
import gromacs.setup

from .datafiles import datafile


@pytest.mark.xfail(
    gromacs.release.startswith("2020.6"),
    reason="pdb2gmx 2020.6 fails to build the TIP4P waters",
)
def test_trj_compact_main(tmpdir):
    pdb = datafile("1ake_A.pdb")
    top = tmpdir.mkdir("top")
    mdpfile = "simple.mdp"
    tprfile = "simple.tpr"
    outfile = "compact.pdb"
    with top.as_cwd():
        f = gromacs.setup.topology(struct=pdb, ff="oplsaa", water="tip4p")
        with open(mdpfile, "w") as mdp:
            mdp.write("; empty mdp file\nrcoulomb = 1\nrvdw = 1\nrlist = 1\n")
        gromacs.grompp(f=mdpfile, o=tprfile, c=f["struct"], p=f["top"])
        gromacs.setup.trj_compact_main(
            s=tprfile, f=f["struct"], o=outfile, input=("protein", "system")
        )
        assert os.path.exists(outfile)


@pytest.fixture(scope="session")
def topology(tmpdir_factory, struct=datafile("1ake_A_protein.pdb")):
    # note: use protein-only input 1ake_A_protein.pdb because solvation fails
    #       if crystal waters are included (in 1ake_A.pdb)
    TMPDIR = tmpdir_factory.mktemp("1ake")
    with TMPDIR.as_cwd():
        topol_args = gromacs.setup.topology(struct=struct, ff="oplsaa", water="tip4p")
    return TMPDIR, topol_args


@pytest.fixture(scope="session")
def solvate(topology):
    TMPDIR, topol_args = topology
    with TMPDIR.as_cwd():
        solvate_args = gromacs.setup.solvate(
            concentration=0.15, water="tip4p", **topol_args
        )
    return TMPDIR, solvate_args


@pytest.fixture(scope="session")
def energy_minimize(solvate, low_performance):
    # run short energy minimization with cheapest minimizer
    TMPDIR, solvate_args = solvate
    nt = 2 if low_performance else 0
    with TMPDIR.as_cwd():
        em_args = gromacs.setup.energy_minimize(
            mdrun_args={"nt": nt},
            integrator="steep",
            emtol=5000,
            maxwarn=1,
            **solvate_args
        )
    return TMPDIR, em_args


def test_topology(topology):
    TMPDIR, topol_args = topology
    top = topol_args["top"]
    struct = topol_args["struct"]
    posres = topol_args["posres"]
    assert os.path.exists(top)
    assert os.path.exists(struct)
    assert os.path.exists(posres)
    # add more tests for content of files!


def test_solvate(solvate):
    TMPDIR, solvate_args = solvate

    assert_almost_equal(solvate_args["qtot"], 0.0)
    assert os.path.exists(solvate_args["struct"])
    assert os.path.exists(solvate_args["ndx"])
    assert solvate_args["mainselection"] == '"Protein"'
    # add more tests for content of files!


def test_energy_minimize(energy_minimize):
    TMPDIR, em_args = energy_minimize
    assert os.path.exists(em_args["struct"])
    assert os.path.exists(em_args["top"])
    assert em_args["mainselection"] == '"Protein"'
    # add more tests for content of files!


def test_energy_minimize_custom_mdp(
    solvate, low_performance, mdp=datafile("custom_em.mdp")
):
    TMPDIR, solvate_args = solvate
    nt = 2 if low_performance else 0
    with TMPDIR.as_cwd():
        try:
            em_args = gromacs.setup.energy_minimize(
                mdrun_args={"nt": nt}, mdp=mdp, emtol=5000, **solvate_args
            )
        except gromacs.exceptions.GromacsError as err:
            # sometimes the em does not converge at all, e.g. 5.02988e+04 on atom 3277;
            # (happens on Travis Linux with Gromacs 4.6.5 but not locally or on Travis OSX) so we
            # re-run with a ridiculous tolerance so that we can at least test that the whole
            # function can run to completion
            em_args = gromacs.setup.energy_minimize(
                mdrun_args={"nt": nt}, mdp=mdp, emtol=6e4, **solvate_args
            )
    assert os.path.exists(em_args["struct"])
    assert os.path.exists(em_args["top"])
    assert em_args["mainselection"] == '"Protein"'
    # add more tests for content of files!



# -*- coding: utf-8 -*-
# GromacsWrapper
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import six

import os.path
import pytest

import gromacs


# use 'ls' as command and only use common BSD/GNU options
@pytest.fixture
def command(command="ls", args=("-a", "-1"), options={"l": True, "v": True}):
    Command_cls = type(
        command.capitalize(), (gromacs.core.Command,), {"command_name": command}
    )
    return Command_cls(*args, **options)


class TestCommand(object):
    def test_run_default(self, command):
        rc, out, err = command()
        assert rc == 0
        assert not out
        assert not err

    def test_run_with_args(self, command):
        rc, out, err = command("-d", os.path.curdir, b=True, F=True)
        assert rc == 0
        assert not out
        assert not err

    def test_run_capture_stdout(self, command):
        rc, out, err = command(stdout=False)
        assert rc == 0
        assert out
        assert not err

    def test_run_capture_stderr(self, command):
        rc, out, err = command("/this_does_not_exist_Foo_Bar", stderr=False)
        assert rc > 0
        assert not out
        assert err

    @pytest.mark.parametrize(
        "inp",
        (
            "not_used",
            ("not", "used"),
            ("unicode", u"Ångström", u"Planck_constant_over_two_π__ℏ"),
        ),
    )
    def test_run_with_input(self, command, inp):
        rc, out, err = command(stdout=False, stderr=False, input=inp)
        assert rc == 0
        assert out
        assert not err

    @pytest.mark.parametrize(
        "inp",
        (
            "not_used",
            ("not", "used"),
            ("unicode", u"Ångström", u"Planck_constant_over_two_π__ℏ"),
        ),
    )
    def test_Popen_with_input(self, command, inp):
        po = command.Popen(stdout=False, stderr=False, input=inp)
        inp_string = "\n".join(gromacs.utilities.asiterable(inp)) + "\n"
        if six.PY2:
            assert po.input == inp_string.encode("utf-8")
        else:
            assert po.input == inp_string

    def test_help_short(self, command, capsys):
        command.help()
        captured = capsys.readouterr()
        assert command.command_name in captured.out

    def test_help_long(self, command, capsys):
        command.help(long=True)
        captured = capsys.readouterr()
        assert command.command_name in captured.out
        assert gromacs.core.Command.__call__.__doc__ in captured.out



# -*- coding: utf-8 -*-
# GromacsWrapper
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import pytest
from numpy.testing import assert_equal, assert_almost_equal

from gromacs.fileformats import convert


@pytest.mark.parametrize(
    "s,expected",
    [
        (100, 100),
        ("Jabberwock", "Jabberwock"),
        (u"Ångström", u"Ångström"),
    ],
)
def test_to_unicode(s, expected):
    output = convert.to_unicode(s)
    assert output == expected


class TestAutoconverter(object):
    def _convert(self, s, **kwargs):
        ac = convert.Autoconverter(**kwargs)
        assert ac.active is True
        return ac.convert(s)

    @pytest.mark.parametrize(
        "s,expected",
        [
            ("foo bar 22  boing ---", ("foo", "bar", 22, "boing", None)),
            ("1 2 3 4", (1, 2, 3, 4)),
            ("1    2  3 4", (1, 2, 3, 4)),
            ("True x X yes Present", (True, True, True, True, True)),
            ("False no - None none", (False, False, False, False, False)),
        ],
    )
    @pytest.mark.parametrize("sep", (True, None))
    def test_convert_default(self, s, expected, sep):
        output = self._convert(s, sep=sep)
        assert_equal(output, expected)

    @pytest.mark.parametrize(
        "s,expected",
        [
            ("1,2,3,4", (1, 2, 3, 4)),
            ("1 2,3,4", ("1 2", 3, 4)),
        ],
    )
    def test_convert_default_sep(self, s, expected, sep=","):
        output = self._convert(s, sep=sep)
        assert_equal(output, expected)

    @pytest.mark.parametrize(
        "s,expected",
        [
            ("2.71213 3.14", (2.71213, 3.14)),
            ("1000 -234 987654", (1000, -234, 987654)),
        ],
    )
    @pytest.mark.parametrize("sep", (True, None))
    def test_convert_numbers(self, s, expected, sep):
        output = self._convert(s, sep=sep)
        assert_almost_equal(output, expected)



from __future__ import division, absolute_import, print_function

import pytest
from numpy.testing import assert_almost_equal, assert_equal

import numpy as np

import gromacs.fileformats.xpm
from gromacs.formats import XPM

from ..datafiles import datafile


@pytest.fixture
def ssfile():
    return datafile("fileformats/ss.xpm.bz2")


@pytest.fixture
def xpm(ssfile):
    return XPM(filename=ssfile)


@pytest.fixture
def xpm_df(xpm):
    return xpm.to_df()


class TestXPM(object):
    def _run_tests(self, x):
        assert_equal(x.array.shape, (500, 769))
        assert_equal(x.xvalues, np.arange(0, 500))
        assert_equal(x.yvalues, np.arange(1, 770))
        assert_equal((x.array == "A-Helix").sum(), 292829)

    def test_constructor(self, xpm):
        self._run_tests(xpm)

    def test_read(self, ssfile):
        x = XPM()
        x.read(ssfile)
        self._run_tests(x)

    def test_reversed_by_default(self, xpm):
        assert xpm.reverse

    def test_to_pd(self, xpm_df):
        assert_equal(xpm_df.shape, (500, 770))

    def test_to_pd_types(self, xpm_df):
        time = xpm_df["Time"]
        assert len(time) == 500
        assert time.dtype == np.dtype("int64")  # true for this file



# GromacsWrapper: xpm.py
# Copyright (c) 2012 Oliver Beckstein <orbeckst@gmail.com>
# Copyright (c) 2010 Tsjerk Wassenaar <tsjerkw@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.
"""
Gromacs XPM file format
=======================

Gromacs stores matrix data in the xpm file format. This implementation
of a Python reader is based on Tsjerk Wassenaar's post to gmx-users
`numerical matrix from xpm file`_ (Mon Oct 4 13:05:26 CEST 2010). This
version returns a NumPy array and can guess an appropriate dtype for
the array.

.. _numerical matrix from xpm file:
   http://lists.gromacs.org/pipermail/gmx-users/2010-October/054557.html

Classes
-------

.. autoclass:: XPM
   :members:

   .. attribute:: xvalues

      Values of on the x-axis, extracted from the xpm file.

   .. attribute:: yvalues

      Values of on the y-axis, extracted from the xpm file. These are
      in the same order as the rows in the xpm matrix. If *reverse* =
      ``False`` then this is typically a *descending* list of numbers
      (highest to lowest residue number, index number, etc). For
      *reverse* = ``True`` it is resorted accordingly.



Example: Analysing H-bonds
--------------------------

Run :func:`gromacs.g_hbond` to produce the existence map (and the log
file for the atoms involved in the bonds; the ndx file is also
useful)::

  gromacs.g_hbond(s=TPR, f=XTC, g="hbond.log", hbm="hb.xpm", hbn="hb.ndx")

Load the XPM::

  hb = XPM("hb.xpm", reverse=True)

Calculate the fraction of time that each H-bond existed::

  hb_fraction = hb.array.mean(axis=0)

Get the descriptions of the bonds::

  desc = [line.strip() for line in open("hbond.log") if not line.startswith('#')]

.. Note::

   It is important that ``reverse=True`` is set so that the rows in
   the xpm matrix are brought in the same order as the H-bond
   labels.

Show the results::

  print "\\n".join(["%-40s %4.1f%%" % p for p in zip(desc, 100*hb_fraction)])

"""

from __future__ import absolute_import, with_statement

from six.moves import range

import os, errno
import re
import warnings

import numpy

from ..exceptions import ParseError, AutoCorrectionWarning
from .. import utilities
from .convert import Autoconverter

import logging


class XPM(utilities.FileUtils):
    """Class to make a Gromacs XPM matrix available as a NumPy :class:`numpy.ndarray`.

    The data is available in the attribute :attr:`XPM.array`.

    .. Note::

       By default, the rows (2nd dimension) in the :attr:`XPM.array`
       are re-ordered so that row 0 (i.e. ``array[:,0]`` corresponds
       to the first residue/hydrogen bond/etc. The original xpm matrix
       is obtained for *reverse* = ``False``. The :class:`XPM` reader
       always reorders the :attr:`XPM.yvalues` (obtained from the xpm
       file) to match the order of the rows.

    """

    default_extension = "xpm"
    logger = logging.getLogger("gromacs.formats.XPM")
    #: compiled regular expression to parse the colors in the xpm file::
    #:
    #:   static char *gromacs_xpm[] = {
    #:   "14327 9   2 1",
    #:   "   c #FFFFFF " /* "None" */,
    #:   "o  c #FF0000 " /* "Present" */,
    #:
    #: Matches are named "symbol", "color" (hex string), and "value". "value"
    #: is typically autoconverted to appropriate values with
    #: :class:`gromacs.fileformats.convert.Autoconverter`.
    #: The symbol is matched as a `printable ASCII character`_ in the range
    #: 0x20 (space) to 0x7E (~).
    #:
    #: .. _`printable ASCII character`: http://www.danshort.com/ASCIImap/indexhex.htm
    COLOUR = re.compile(
        """\
            ^.*"                   # start with quotation mark
            (?P<symbol>[\x20-\x7E])# printable ASCII symbol used in the actual pixmap: 'space' to '~'
            \s+                    # white-space separated
            c\s+                   # 'c' to prefix colour??
            (?P<color>\#[0-9A-F]+) # colour as hex string (always??)
            \s*"                   # close with quotes
            \s*/\*\s*              # white space then opening C-comment /*
            "                      # start new string
            (?P<value>.*)          # description/value as free form string
            "                      # ... terminated by quotes
            """,
        re.VERBOSE,
    )

    def __init__(self, filename=None, **kwargs):
        """Initialize xpm structure.

        :Arguments:
          *filename*
              read from xpm file directly
          *autoconvert*
              try to guess the type of the output array from the
              colour legend [``True``]
          *reverse*
              reverse rows (2nd dimension): re-orders the rows so that
              the first row corresponds e.g. to the first residue or
              first H-bonds and not the last) [``True``]
        """
        self.autoconvert = kwargs.pop("autoconvert", True)
        self.reverse = kwargs.pop("reverse", True)
        self.__array = None
        super(XPM, self).__init__(
            **kwargs
        )  # can use kwargs to set dict! (but no sanity checks!)

        if filename is not None:
            self._init_filename(filename)
            self.read(filename)

    def to_df(self):
        import pandas as pd

        # Add Time to the data as column
        data = numpy.vstack((self.xvalues, self.array.T)).T

        # Column names are resids
        df = pd.DataFrame(data, columns=["Time"] + list(self.yvalues))

        # Converts Time to a numeric type
        df["Time"] = pd.to_numeric(df["Time"])
        return df

    @property
    def array(self):
        """XPM matrix as a :class:`numpy.ndarray`.

        The attribute itself cannot be assigned a different array but
        the contents of the array can be modified.
        """
        return self.__array

    def read(self, filename=None):
        """Read and parse mdp file *filename*."""
        self._init_filename(filename)
        self.parse()

    def parse(self):
        """Parse the xpm file and populate :attr:`XPM.array`."""
        with utilities.openany(self.real_filename) as xpm:
            # Read in lines until we find the start of the array
            meta = [xpm.readline()]
            while not meta[-1].startswith("static char *gromacs_xpm[]"):
                meta.append(xpm.readline())

            # The next line will contain the dimensions of the array
            dim = xpm.readline()
            # There are four integers surrounded by quotes
            # nx: points along x, ny: points along y, nc: ?, nb: stride x
            nx, ny, nc, nb = [int(i) for i in self.unquote(dim).split()]

            # The next dim[2] lines contain the color definitions
            # Each pixel is encoded by dim[3] bytes, and a comment
            # at the end of the line contains the corresponding value
            colors = dict([self.col(xpm.readline()) for i in range(nc)])

            if self.autoconvert:
                autoconverter = Autoconverter(mode="singlet")
                for symbol, value in colors.items():
                    colors[symbol] = autoconverter.convert(value)
                self.logger.debug("Autoconverted colours: %r", colors)

            # make an array containing all possible values and let numpy figure out the dtype
            dtype = numpy.array(colors.values()).dtype
            self.logger.debug("Guessed array type: %s", dtype.name)

            # pre-allocate array
            data = numpy.zeros((int(nx / nb), ny), dtype=dtype)

            self.logger.debug(
                "dimensions: NX=%d NY=%d strideX=%d (NC=%d) --> (%d, %d)",
                nx,
                ny,
                nb,
                nc,
                nx / nb,
                ny,
            )

            iy = 0
            xval = []
            yval = []
            autoconverter = Autoconverter(mode="singlet")
            for line in xpm:
                if line.startswith("/*"):
                    # lines '/* x-axis:' ... and '/* y-axis:' contain the
                    # values of x and y coordinates
                    s = self.uncomment(line).strip()
                    if s.startswith("x-axis:"):
                        xval.extend([autoconverter.convert(x) for x in s[7:].split()])
                    elif s.startswith("y-axis:"):
                        yval.extend([autoconverter.convert(y) for y in s[7:].split()])
                    continue
                s = self.unquote(line)
                # Joao M. Damas <jmdamas@itqb.unl.pt> suggests on gmx-users (24 Oct 2014)
                # that the next line should read:
                #
                #  data[:, iy]  =  [colors[j[k:k+nb]] for k in range(0,nx*nb,nb)]
                #
                # "if one is using higher -nlevels for the .xpm construction (in g_rms, for example)"
                # However, without a test case I am not eager to change it right away so in
                # case some faulty behavior is discovered with the XPM reader then this comment
                # might be helpful. --- Oliver 2014-10-25
                data[:, iy] = [colors[s[k : k + nb]] for k in range(0, nx, nb)]
                self.logger.debug(
                    "read row %d with %d columns: '%s....%s'",
                    iy,
                    data.shape[0],
                    s[:4],
                    s[-4:],
                )
                iy += 1  # for next row

        self.xvalues = numpy.array(xval)
        if self.reverse:
            self.logger.debug("reversed row order, reverse=%r", self.reverse)
            self.__array = data[:, ::-1]
            self.yvalues = numpy.array(yval)
        else:
            self.__array = data
            self.yvalues = numpy.array(yval)[::-1]  # must reverse y-values to match!

    @staticmethod
    def unquote(s):
        """Return string *s* with quotes ``"`` removed."""
        return s[1 + s.find('"') : s.rfind('"')]

    @staticmethod
    def uncomment(s):
        """Return string *s* with C-style comments ``/*`` ... ``*/`` removed."""
        return s[2 + s.find("/*") : s.rfind("*/")]

    def col(self, c):
        """Parse colour specification"""
        m = self.COLOUR.search(c)
        if not m:
            self.logger.fatal("Cannot parse colour specification %r.", c)
            raise ParseError(
                "XPM reader: Cannot parse colour specification {0!r}.".format(c)
            )
        value = m.group("value")
        color = m.group("symbol")
        self.logger.debug("%s: %s %s\n", c.strip(), color, value)
        return color, value
